"""
Provide no-op caching.
"""

from __future__ import annotations

from contextlib import asynccontextmanager
from typing import TYPE_CHECKING, Any, Self, final, override

from betty.cache import Cache, CacheItem, CacheItemValueSetter
from betty.typing import threadsafe

if TYPE_CHECKING:
    from collections.abc import AsyncIterator


@final
@threadsafe
class NoOpCache(Cache[Any]):
    """
    Provide a cache that does nothing.
    """

    @override
    def with_scope(self, scope: str, /) -> Self:
        return self

    async def _setter(self, value: Any) -> None:
        pass

    @override
    async def has(self, cache_item_id: str, /) -> bool:
        return False

    @override
    @asynccontextmanager
    async def hasset(
        self, cache_item_id: str, /
    ) -> AsyncIterator[CacheItemValueSetter[Any] | None]:
        yield self._setter
        return

    @override
    async def get(self, cache_item_id: str, /) -> CacheItem[Any] | None:
        return None

    @override
    async def set(
        self,
        cache_item_id: str,
        value: Any,
        *,
        modified: int | float | None = None,
    ) -> None:
        return

    @override
    @asynccontextmanager
    async def getset(
        self, cache_item_id: str, /
    ) -> AsyncIterator[CacheItemValueSetter[Any] | CacheItem[Any]]:
        yield self._setter
        return

    @override
    async def delete(self, cache_item_id: str, /) -> None:
        return

    @override
    async def clear(self) -> None:
        return



"""
Test utilities for :py:mod:`betty.user`.
"""

import logging
import sys
from collections.abc import AsyncIterator, Collection, Iterable, MutableSequence
from contextlib import asynccontextmanager
from typing import overload, override

from betty.assertion import Assertion
from betty.locale.localizable import ResolvableLocalizable
from betty.locale.localize import DEFAULT_LOCALIZER, resolve_localized
from betty.progress import Progress
from betty.progress.no_op import NoOpProgress
from betty.typing import Void, VoidType
from betty.user import User, UserTimeoutError, Verbosity


class StaticUser(User):  # pragma: no cover
    """
    A static user with predefined responses.
    """

    verbosity = Verbosity.DEFAULT

    def __init__(
        self,
        *,
        confirmations: Iterable[bool | None] | None = None,
        inputs: Iterable[str | None] | None = None,
    ):
        self._confirmations = iter([] if confirmations is None else confirmations)
        self._inputs = iter([] if inputs is None else inputs)
        self._messages_exception: MutableSequence[BaseException] = []
        self._messages_error: MutableSequence[ResolvableLocalizable] = []
        self._messages_warning: MutableSequence[ResolvableLocalizable] = []
        self._messages_information: MutableSequence[ResolvableLocalizable] = []
        self._messages_information_details: MutableSequence[ResolvableLocalizable] = []
        self._messages_debug: MutableSequence[ResolvableLocalizable] = []
        self._messages_log: MutableSequence[logging.LogRecord] = []
        self._log_formatter = logging.Formatter()

    @override
    async def set_verbosity(self, verbosity: Verbosity, /) -> None:
        self.verbosity = verbosity

    def _format_fragments(self, fragments: str | Iterable[str]) -> str:
        if isinstance(fragments, str):
            fragments = [fragments]
        return ", ".join(f'"{fragment}"' for fragment in fragments)

    def _message_contains(self, message: str, fragments: Iterable[str]) -> bool:
        return all(fragment in message for fragment in fragments)

    def _assert_message(
        self,
        fragments: str | Iterable[str],
        message_type: str,
        messages: Collection[str],
    ) -> None:
        if isinstance(fragments, str):
            fragments = [fragments]
        for message in messages:
            if self._message_contains(message, fragments):
                return
        raise AssertionError(
            f'Failed asserting that a(n) "{message_type}" message was sent containing the fragment(s) {self._format_fragments(fragments)}.'
        )

    def _assert_localizable_message(
        self, fragments: str | Iterable[str], message_type: str
    ) -> None:
        self._assert_message(
            fragments,
            message_type,
            [
                resolve_localized(message, localizer=DEFAULT_LOCALIZER)
                for message in getattr(self, f"_messages_{message_type}")
            ],
        )

    def assert_message_exception(self, fragments: str | Iterable[str]) -> None:
        """
        Assert that an exception message was sent.
        """
        self._assert_message(
            fragments, "exception", list(map(str, self._messages_exception))
        )

    def assert_message_error(self, fragments: str | Iterable[str]) -> None:
        """
        Assert that an error message was sent.
        """
        self._assert_localizable_message(fragments, "error")

    def assert_message_warning(self, fragments: str | Iterable[str]) -> None:
        """
        Assert that a warning message was sent.
        """
        self._assert_localizable_message(fragments, "warning")

    def assert_message_information(self, fragments: str | Iterable[str]) -> None:
        """
        Assert that an information message was sent.
        """
        self._assert_localizable_message(fragments, "information")

    def assert_message_information_details(
        self, fragments: str | Iterable[str]
    ) -> None:
        """
        Assert that a detailed information message was sent.
        """
        self._assert_localizable_message(fragments, "information_details")

    def assert_message_debug(self, fragments: str | Iterable[str]) -> None:
        """
        Assert that a debug message was sent.
        """
        self._assert_localizable_message(fragments, "debug")

    def assert_message_log(self, fragments: str | Iterable[str]) -> None:
        """
        Assert that a log message was sent.
        """
        self._assert_message(
            fragments, "log", list(map(self._log_formatter.format, self._messages_log))
        )

    def _assert_not_message(
        self,
        fragments: str | Iterable[str],
        message_type: str,
        messages: Collection[str],
    ) -> None:
        if isinstance(fragments, str):
            fragments = [fragments]
        for message in messages:
            if self._message_contains(message, fragments):
                raise AssertionError(
                    f'Failed asserting that a(n) "{message_type}" message was sent containing the fragment(s) {self._format_fragments(fragments)}.'
                )

    def _assert_not_localizable_message(
        self, fragments: str | Iterable[str], message_type: str
    ) -> None:
        self._assert_not_message(
            fragments,
            message_type,
            [
                resolve_localized(message, localizer=DEFAULT_LOCALIZER)
                for message in getattr(self, f"_messages_{message_type}")
            ],
        )

    def assert_not_message_exception(self, fragments: str | Iterable[str]) -> None:
        """
        Assert that no exception message was sent.
        """
        self._assert_not_message(
            fragments, "exception", list(map(str, self._messages_exception))
        )

    def assert_not_message_error(self, fragments: str | Iterable[str]) -> None:
        """
        Assert that no error message was sent.
        """
        self._assert_not_localizable_message(fragments, "error")

    def assert_not_message_warning(self, fragments: str | Iterable[str]) -> None:
        """
        Assert that no warning message was sent.
        """
        self._assert_not_localizable_message(fragments, "warning")

    def assert_not_message_information(self, fragments: str | Iterable[str]) -> None:
        """
        Assert that no information message was sent.
        """
        self._assert_not_localizable_message(fragments, "information")

    def assert_not_message_information_details(
        self, fragments: str | Iterable[str]
    ) -> None:
        """
        Assert that no detailed information message was sent.
        """
        self._assert_not_localizable_message(fragments, "information_details")

    def assert_not_message_debug(self, fragments: str | Iterable[str]) -> None:
        """
        Assert that no debug message was sent.
        """
        self._assert_not_localizable_message(fragments, "debug")

    def assert_not_message_log(self, fragments: str | Iterable[str]) -> None:
        """
        Assert that no log message was sent.
        """
        self._assert_not_message(
            fragments, "log", list(map(self._log_formatter.format, self._messages_log))
        )

    @override
    async def message_exception(self) -> None:
        exception = sys.exception()
        assert exception
        self._messages_exception.append(exception)

    @override
    async def message_error(self, message: ResolvableLocalizable, /) -> None:
        self._messages_error.append(message)

    @override
    async def message_warning(self, message: ResolvableLocalizable, /) -> None:
        self._messages_warning.append(message)

    @override
    async def message_information(self, message: ResolvableLocalizable, /) -> None:
        self._messages_information.append(message)

    @override
    async def message_information_details(
        self, message: ResolvableLocalizable, /
    ) -> None:
        self._messages_information_details.append(message)

    @override
    async def message_debug(self, message: ResolvableLocalizable, /) -> None:
        self._messages_debug.append(message)

    @override
    async def message_log(self, message: logging.LogRecord, /) -> None:
        self._messages_log.append(message)

    @override
    @asynccontextmanager
    async def message_progress(
        self, message: ResolvableLocalizable, /
    ) -> AsyncIterator[Progress]:
        yield NoOpProgress()

    @override
    async def ask_confirmation(
        self, statement: ResolvableLocalizable, *, default: bool = False
    ) -> bool:
        confirmation = next(self._confirmations)
        if confirmation is None:
            return default
        return confirmation

    @overload
    async def ask_input(
        self,
        question: ResolvableLocalizable,
        *,
        default: str | VoidType = Void,
    ) -> str:
        pass

    @overload
    async def ask_input[T](
        self,
        question: ResolvableLocalizable,
        *,
        assertion: Assertion[str, T],
        default: str | VoidType = Void,
    ) -> T:
        pass

    @override
    async def ask_input[T](
        self,
        question: ResolvableLocalizable,
        *,
        assertion: Assertion[str, T] | None = None,
        default: str | T | VoidType = Void,
    ) -> str | T:
        value = next(self._inputs)
        if value is None:
            if default is Void:
                raise UserTimeoutError(
                    "Neither a predefined response nor a call default were provided."
                )
            return default
        if assertion is None:
            return value
        return assertion(value)



from __future__ import annotations  # noqa: D100

import platform
import sys
from importlib import metadata
from typing import TYPE_CHECKING, Self, final, override

from rich.table import Table

from betty import about
from betty.console.command import Command, CommandDefinition, CommandFunction
from betty.console.project import add_project_argument
from betty.definition.human_facing import HumanFacingDefinition
from betty.locale.localizable.gettext import _
from betty.rich.user import RichUser
from betty.service.factory import Manufacturable
from betty.service.requirement.app import require_app

if TYPE_CHECKING:
    import argparse
    from collections.abc import MutableSequence

    from betty.app import App
    from betty.project import Project


@final
@CommandDefinition(
    "about", label=_("Output information about Betty, and optionally your project")
)
class About(Manufacturable, Command):
    """
    .. plugin:: command:about.
    """

    _KEY_STYLE = "cyan"

    def __init__(self, app: App):
        self._app = app

    @override
    @classmethod
    @require_app
    async def new(cls, app: App, /) -> Self:
        return cls(app)

    @override
    async def configure(self, parser: argparse.ArgumentParser) -> CommandFunction:
        return await add_project_argument(
            parser, self._command_function, self._app, required=False
        )

    async def _command_function(self, project: Project | None) -> None:
        user = self._app.user
        assert isinstance(user, RichUser)
        try:
            if project:
                await project.bootstrap()
                await self._about_project(user, project)
            await self._about_plugins(user, project)
            await self._about_python_packages(user)
            await self._about_system(user)
        finally:
            if project:
                await project.shutdown()

    async def _about_project(self, user: RichUser, project: Project) -> None:
        about_project = Table(
            title=user.localizer._("Your project at {file}").format(
                file=str(project.configuration_file.parent)
            ),
            show_header=False,
        )
        about_project.add_column("", style=self._KEY_STYLE)
        about_project.add_column("")
        about_project.add_row(
            user.localizer._("Configuration file"),
            str(project.configuration_file),
        )
        about_project.add_row(
            user.localizer._("Assets directory"),
            str(project.assets_directory),
        )
        about_project.add_row(
            user.localizer._("Output directory"),
            str(project.output_directory),
        )
        user.console.print(about_project)

    async def _about_plugins(self, user: RichUser, project: Project | None) -> None:
        services = self._app if project is None else project
        about_plugins = Table(title=user.localizer._("Plugins"))
        about_plugins.add_column(user.localizer._("Type"), style=self._KEY_STYLE)
        about_plugins.add_column(user.localizer._("ID"))
        about_plugins.add_column(user.localizer._("Label"))
        for plugin_manager in sorted(
            services.plugins,
            key=lambda plugin_type: plugin_type.type.type().label.localize(
                user.localizer
            ),
        ):
            for index, plugin in enumerate(
                sorted([x async for x in plugin_manager], key=lambda plugin: plugin.id)
            ):
                first_column = (
                    plugin_manager.type.type().label.localize(user.localizer)
                    if index == 0
                    else ""
                )
                third_column_lines: MutableSequence[str] = []
                if isinstance(plugin, HumanFacingDefinition):
                    third_column_lines.append(plugin.label.localize(user.localizer))
                about_plugins.add_row(
                    first_column,
                    plugin.id,
                    "\n".join(third_column_lines),
                )
        user.console.print(about_plugins)
        if project is None:
            user.console.print(
                "[yellow]"
                + user.localizer._(
                    "More plugins may be available when running this command with --project."
                )
            )

    async def _about_system(self, user: RichUser) -> None:
        about_system = Table(title=user.localizer._("System"), show_header=False)
        about_system.add_column("", style=self._KEY_STYLE)
        about_system.add_column("")
        about_system.add_row("Betty", about.VERSION_LABEL)
        about_system.add_row(user.localizer._("Operating system"), platform.platform())
        about_system.add_row("Python", sys.version)
        user.console.print(about_system)

    async def _about_python_packages(self, user: RichUser) -> None:
        about_python_packages = Table(title=user.localizer._("Python packages"))
        about_python_packages.add_column(
            user.localizer._("Package"), style=self._KEY_STYLE
        )
        about_python_packages.add_column(user.localizer._("Version"))
        for x in sorted(
            metadata.distributions(),
            key=lambda x: x.metadata["Name"].lower(),
        ):
            about_python_packages.add_row(x.metadata["Name"], x.version)
        user.console.print(about_python_packages)



"""Integrate Betty with `Wikipedia <https://wikipedia.org>`_."""

from __future__ import annotations

from asyncio import gather
from pathlib import Path
from typing import TYPE_CHECKING, Self, final, override

from jinja2 import pass_context

from betty.copyright_notice import CopyrightNoticeDefinition
from betty.extension import Extension, ExtensionDefinition
from betty.extension.wiki.data import WikiConfiguration
from betty.extension.wiki.jobs import PopulateEntity
from betty.jinja import Filters, Globals, JinjaProvider, context_localizer
from betty.locale import negotiate_locale, resolve_locale
from betty.locale.localizable.gettext import _
from betty.project import Project
from betty.project.load import PostLoader
from betty.service.factory import DataManufacturable, Manufacturable
from betty.service.provider import service
from betty.service.requirement.project import require_project
from betty.wiki import NotAPageError, parse_page_url
from betty.wiki import populator as populator_api
from betty.wiki.client import Client, ClientError, Summary

if TYPE_CHECKING:
    from collections.abc import Iterable

    from babel import Locale
    from jinja2.runtime import Context

    from betty.ancestry.link import Link
    from betty.copyright_notice import CopyrightNotice
    from betty.job.scheduler import Scheduler


@final
@ExtensionDefinition(
    "wiki",
    label="Wiki",
    description=_(
        "Enrich your ancestry with information from Wikipedia and Wikimedia Commons"
    ),
    assets_directory=Path(__file__).parent / "assets",
)
class Wiki(
    PostLoader,
    DataManufacturable[WikiConfiguration],
    Manufacturable,
    JinjaProvider,
    Extension[Project],
):
    """
    .. plugin:: extension:wiki.

    Links
    -----
    For the extension to know where to look for information, simply add a single link to a human-readable Wikipedia page to that entity's links.

    Ancestry enrichment
    -------------------
    The extension will attempt the following for any entity that has a Wikipedia link:

    - for places, add coordinates if a place has none already
    - for any entity, add additional links to the translations of the given Wikipedia page
    - for any entity that has files, add the primary image of the linked Wikipedia page

    Templating
    ----------

    Globals
    ^^^^^^^

    ``wikipedia_contributors_copyright_notice`` (:py:class:`betty.copyright_notice.copyright_notices.WikipediaContributors`)
        The copyright notice plugin instance for Wikipedia contributors.

    Filters
    ^^^^^^^

    - :py:meth:`wikipedia_summary <betty.extension.wiki.Wiki.filter_wikipedia_summary_links>`

    """

    def __init__(
        self,
        *,
        project: Project,
        wikipedia_contributors_copyright_notice: CopyrightNotice,
        populate_images: bool | None = None,
    ):
        super().__init__()
        self._project = project
        self._populate_images = True if populate_images is None else populate_images
        self._wikipedia_contributors_copyright_notice = (
            wikipedia_contributors_copyright_notice
        )

    @override
    @classmethod
    def new_data_cls(cls) -> type[WikiConfiguration]:
        return WikiConfiguration

    @override
    @classmethod
    @require_project
    async def new(
        cls, project: Project, data: WikiConfiguration | None = None, /
    ) -> Self:
        copyright_notices = project.plugins[CopyrightNoticeDefinition]
        return cls(
            populate_images=None if data is None else data.populate_images,
            project=project,
            wikipedia_contributors_copyright_notice=await project.factory.new(
                (await copyright_notices["wikipedia-contributors"]).cls
            ),
        )

    @override
    async def post_load(self, scheduler: Scheduler) -> None:
        await scheduler.add(
            *(
                PopulateEntity(entity, project=self._project)
                for entity in self._project.ancestry
            )
        )

    @service
    async def client(self) -> Client:
        """
        The API client.
        """
        return Client(
            download_directory_path=self._project.app.binary_file_cache.with_scope(
                "wiki-client"
            ).path,
            http_client=await self._project.app.http_client,
            user=self._project.app.user,
        )

    @service
    async def populator(self) -> populator_api.Populator:
        """
        The ancestry populator.
        """
        return populator_api.Populator(
            self._project.ancestry,
            list(self._project.configuration.locales.keys()),
            await self._project.localizers,
            await self.client,
            self._wikipedia_contributors_copyright_notice,
            user=self._project.app.user,
        )

    @override
    @property
    def globals(self) -> Globals:
        return {
            "wikipedia_contributors_copyright_notice": self._wikipedia_contributors_copyright_notice
        }

    @override
    @property
    def filters(self) -> Filters:
        return {
            "wikipedia_summary": self.filter_wikipedia_summary_links,
        }

    @pass_context
    async def filter_wikipedia_summary_links(
        self, context: Context, links: Iterable[Link]
    ) -> Iterable[Summary]:
        """
        Given a sequence of links, return any Wikipedia summaries for them.
        """
        return filter(
            None,
            await gather(
                *(
                    self._filter_wikipedia_summary_link(
                        context_localizer(context).locale,
                        link,
                    )
                    for link in links
                )
            ),
        )

    async def _filter_wikipedia_summary_link(
        self, locale: Locale, link: Link
    ) -> Summary | None:
        localizers = await self._project.app.localizers
        try:
            page_language, page_name = parse_page_url(
                link.url.localize(localizers.get(locale))
            )
        except NotAPageError:
            return None
        if (
            negotiate_locale(
                locale, list(filter(None, [resolve_locale(page_language)]))
            )
            is None
        ):
            return None
        try:
            client = await self.client
            return await client.get_summary(page_language, page_name)
        except ClientError:
            return None



"""
JSON schemas for the OpenAPI API.
"""

from typing import final

from betty.classtools import Singleton
from betty.json.schema import Schema
from betty.portable import PortableMapping


@final
class SpecificationSchema(Singleton, Schema):
    """
    The OpenAPI Specification schema.
    """

    _SCHEMA: PortableMapping = {
        "$id": "https://spec.openapis.org/oas/3.1/schema/2022-10-07",
        "$schema": "https://json-schema.org/draft/2020-12/schema",
        "description": "The description of OpenAPI v3.1.x documents without schema validation, as defined by https://spec.openapis.org/oas/v3.1.0",
        "type": "object",
        "properties": {
            "openapi": {"type": "string", "pattern": "^3\\.1\\.\\d+(-.+)?$"},
            "info": {"$ref": "#/$defs/info"},
            "jsonSchemaDialect": {
                "type": "string",
                "format": "uri",
                "default": "https://spec.openapis.org/oas/3.1/dialect/base",
            },
            "servers": {
                "type": "array",
                "items": {"$ref": "#/$defs/server"},
                "default": [{"url": "/"}],
            },
            "paths": {"$ref": "#/$defs/paths"},
            "webhooks": {
                "type": "object",
                "additionalProperties": {"$ref": "#/$defs/path-item-or-reference"},
            },
            "components": {"$ref": "#/$defs/components"},
            "security": {
                "type": "array",
                "items": {"$ref": "#/$defs/security-requirement"},
            },
            "tags": {"type": "array", "items": {"$ref": "#/$defs/tag"}},
            "externalDocs": {"$ref": "#/$defs/external-documentation"},
        },
        "required": ["openapi", "info"],
        "anyOf": [
            {"required": ["paths"]},
            {"required": ["components"]},
            {"required": ["webhooks"]},
        ],
        "$ref": "#/$defs/specification-extensions",
        "unevaluatedProperties": False,
        "$defs": {
            "info": {
                "$comment": "https://spec.openapis.org/oas/v3.1.0#info-object",
                "type": "object",
                "properties": {
                    "title": {"type": "string"},
                    "summary": {"type": "string"},
                    "description": {"type": "string"},
                    "termsOfService": {"type": "string", "format": "uri"},
                    "contact": {"$ref": "#/$defs/contact"},
                    "license": {"$ref": "#/$defs/license"},
                    "version": {"type": "string"},
                },
                "required": ["title", "version"],
                "$ref": "#/$defs/specification-extensions",
                "unevaluatedProperties": False,
            },
            "contact": {
                "$comment": "https://spec.openapis.org/oas/v3.1.0#contact-object",
                "type": "object",
                "properties": {
                    "name": {"type": "string"},
                    "url": {"type": "string", "format": "uri"},
                    "email": {"type": "string", "format": "email"},
                },
                "$ref": "#/$defs/specification-extensions",
                "unevaluatedProperties": False,
            },
            "license": {
                "$comment": "https://spec.openapis.org/oas/v3.1.0#license-object",
                "type": "object",
                "properties": {
                    "name": {"type": "string"},
                    "identifier": {"type": "string"},
                    "url": {"type": "string", "format": "uri"},
                },
                "required": ["name"],
                "dependentSchemas": {"identifier": {"not": {"required": ["url"]}}},
                "$ref": "#/$defs/specification-extensions",
                "unevaluatedProperties": False,
            },
            "server": {
                "$comment": "https://spec.openapis.org/oas/v3.1.0#server-object",
                "type": "object",
                "properties": {
                    "url": {"type": "string", "format": "uri-reference"},
                    "description": {"type": "string"},
                    "variables": {
                        "type": "object",
                        "additionalProperties": {"$ref": "#/$defs/server-variable"},
                    },
                },
                "required": ["url"],
                "$ref": "#/$defs/specification-extensions",
                "unevaluatedProperties": False,
            },
            "server-variable": {
                "$comment": "https://spec.openapis.org/oas/v3.1.0#server-variable-object",
                "type": "object",
                "properties": {
                    "enum": {
                        "type": "array",
                        "items": {"type": "string"},
                        "minItems": 1,
                    },
                    "default": {"type": "string"},
                    "description": {"type": "string"},
                },
                "required": ["default"],
                "$ref": "#/$defs/specification-extensions",
                "unevaluatedProperties": False,
            },
            "components": {
                "$comment": "https://spec.openapis.org/oas/v3.1.0#components-object",
                "type": "object",
                "properties": {
                    "schemas": {
                        "type": "object",
                        "additionalProperties": {"$dynamicRef": "#meta"},
                    },
                    "responses": {
                        "type": "object",
                        "additionalProperties": {
                            "$ref": "#/$defs/response-or-reference"
                        },
                    },
                    "parameters": {
                        "type": "object",
                        "additionalProperties": {
                            "$ref": "#/$defs/parameter-or-reference"
                        },
                    },
                    "examples": {
                        "type": "object",
                        "additionalProperties": {
                            "$ref": "#/$defs/example-or-reference"
                        },
                    },
                    "requestBodies": {
                        "type": "object",
                        "additionalProperties": {
                            "$ref": "#/$defs/request-body-or-reference"
                        },
                    },
                    "headers": {
                        "type": "object",
                        "additionalProperties": {"$ref": "#/$defs/header-or-reference"},
                    },
                    "securitySchemes": {
                        "type": "object",
                        "additionalProperties": {
                            "$ref": "#/$defs/security-scheme-or-reference"
                        },
                    },
                    "links": {
                        "type": "object",
                        "additionalProperties": {"$ref": "#/$defs/link-or-reference"},
                    },
                    "callbacks": {
                        "type": "object",
                        "additionalProperties": {
                            "$ref": "#/$defs/callbacks-or-reference"
                        },
                    },
                    "pathItems": {
                        "type": "object",
                        "additionalProperties": {
                            "$ref": "#/$defs/path-item-or-reference"
                        },
                    },
                },
                "patternProperties": {
                    "^(schemas|responses|parameters|examples|requestBodies|headers|securitySchemes|links|callbacks|pathItems)$": {
                        "$comment": "Enumerating all of the property names in the regex above is necessary for unevaluatedProperties to work as expected",
                        "propertyNames": {"pattern": "^[a-zA-Z0-9._-]+$"},
                    }
                },
                "$ref": "#/$defs/specification-extensions",
                "unevaluatedProperties": False,
            },
            "paths": {
                "$comment": "https://spec.openapis.org/oas/v3.1.0#paths-object",
                "type": "object",
                "patternProperties": {"^/": {"$ref": "#/$defs/path-item"}},
                "$ref": "#/$defs/specification-extensions",
                "unevaluatedProperties": False,
            },
            "path-item": {
                "$comment": "https://spec.openapis.org/oas/v3.1.0#path-item-object",
                "type": "object",
                "properties": {
                    "summary": {"type": "string"},
                    "description": {"type": "string"},
                    "servers": {"type": "array", "items": {"$ref": "#/$defs/server"}},
                    "parameters": {
                        "type": "array",
                        "items": {"$ref": "#/$defs/parameter-or-reference"},
                    },
                    "get": {"$ref": "#/$defs/operation"},
                    "put": {"$ref": "#/$defs/operation"},
                    "post": {"$ref": "#/$defs/operation"},
                    "delete": {"$ref": "#/$defs/operation"},
                    "options": {"$ref": "#/$defs/operation"},
                    "head": {"$ref": "#/$defs/operation"},
                    "patch": {"$ref": "#/$defs/operation"},
                    "trace": {"$ref": "#/$defs/operation"},
                },
                "$ref": "#/$defs/specification-extensions",
                "unevaluatedProperties": False,
            },
            "path-item-or-reference": {
                "if": {"type": "object", "required": ["$ref"]},
                "then": {"$ref": "#/$defs/reference"},
                "else": {"$ref": "#/$defs/path-item"},
            },
            "operation": {
                "$comment": "https://spec.openapis.org/oas/v3.1.0#operation-object",
                "type": "object",
                "properties": {
                    "tags": {"type": "array", "items": {"type": "string"}},
                    "summary": {"type": "string"},
                    "description": {"type": "string"},
                    "externalDocs": {"$ref": "#/$defs/external-documentation"},
                    "operationId": {"type": "string"},
                    "parameters": {
                        "type": "array",
                        "items": {"$ref": "#/$defs/parameter-or-reference"},
                    },
                    "requestBody": {"$ref": "#/$defs/request-body-or-reference"},
                    "responses": {"$ref": "#/$defs/responses"},
                    "callbacks": {
                        "type": "object",
                        "additionalProperties": {
                            "$ref": "#/$defs/callbacks-or-reference"
                        },
                    },
                    "deprecated": {"default": False, "type": "boolean"},
                    "security": {
                        "type": "array",
                        "items": {"$ref": "#/$defs/security-requirement"},
                    },
                    "servers": {"type": "array", "items": {"$ref": "#/$defs/server"}},
                },
                "$ref": "#/$defs/specification-extensions",
                "unevaluatedProperties": False,
            },
            "external-documentation": {
                "$comment": "https://spec.openapis.org/oas/v3.1.0#external-documentation-object",
                "type": "object",
                "properties": {
                    "description": {"type": "string"},
                    "url": {"type": "string", "format": "uri"},
                },
                "required": ["url"],
                "$ref": "#/$defs/specification-extensions",
                "unevaluatedProperties": False,
            },
            "parameter": {
                "$comment": "https://spec.openapis.org/oas/v3.1.0#parameter-object",
                "type": "object",
                "properties": {
                    "name": {"type": "string"},
                    "in": {"enum": ["query", "header", "path", "cookie"]},
                    "description": {"type": "string"},
                    "required": {"default": False, "type": "boolean"},
                    "deprecated": {"default": False, "type": "boolean"},
                    "schema": {"$dynamicRef": "#meta"},
                    "content": {
                        "$ref": "#/$defs/content",
                        "minProperties": 1,
                        "maxProperties": 1,
                    },
                },
                "required": ["name", "in"],
                "oneOf": [{"required": ["schema"]}, {"required": ["content"]}],
                "if": {"properties": {"in": {"const": "query"}}, "required": ["in"]},
                "then": {
                    "properties": {
                        "allowEmptyValue": {"default": False, "type": "boolean"}
                    }
                },
                "dependentSchemas": {
                    "schema": {
                        "properties": {
                            "style": {"type": "string"},
                            "explode": {"type": "boolean"},
                        },
                        "allOf": [
                            {"$ref": "#/$defs/examples"},
                            {
                                "$ref": "#/$defs/parameter/dependentSchemas/schema/$defs/styles-for-path"
                            },
                            {
                                "$ref": "#/$defs/parameter/dependentSchemas/schema/$defs/styles-for-header"
                            },
                            {
                                "$ref": "#/$defs/parameter/dependentSchemas/schema/$defs/styles-for-query"
                            },
                            {
                                "$ref": "#/$defs/parameter/dependentSchemas/schema/$defs/styles-for-cookie"
                            },
                            {
                                "$ref": "#/$defs/parameter/dependentSchemas/schema/$defs/styles-for-form"
                            },
                        ],
                        "$defs": {
                            "styles-for-path": {
                                "if": {
                                    "properties": {"in": {"const": "path"}},
                                    "required": ["in"],
                                },
                                "then": {
                                    "properties": {
                                        "name": {"pattern": "[^/#?]+$"},
                                        "style": {
                                            "default": "simple",
                                            "enum": ["matrix", "label", "simple"],
                                        },
                                        "required": {"const": True},
                                    },
                                    "required": ["required"],
                                },
                            },
                            "styles-for-header": {
                                "if": {
                                    "properties": {"in": {"const": "header"}},
                                    "required": ["in"],
                                },
                                "then": {
                                    "properties": {
                                        "style": {
                                            "default": "simple",
                                            "const": "simple",
                                        }
                                    }
                                },
                            },
                            "styles-for-query": {
                                "if": {
                                    "properties": {"in": {"const": "query"}},
                                    "required": ["in"],
                                },
                                "then": {
                                    "properties": {
                                        "style": {
                                            "default": "form",
                                            "enum": [
                                                "form",
                                                "spaceDelimited",
                                                "pipeDelimited",
                                                "deepObject",
                                            ],
                                        },
                                        "allowReserved": {
                                            "default": False,
                                            "type": "boolean",
                                        },
                                    }
                                },
                            },
                            "styles-for-cookie": {
                                "if": {
                                    "properties": {"in": {"const": "cookie"}},
                                    "required": ["in"],
                                },
                                "then": {
                                    "properties": {
                                        "style": {"default": "form", "const": "form"}
                                    }
                                },
                            },
                            "styles-for-form": {
                                "if": {
                                    "properties": {"style": {"const": "form"}},
                                    "required": ["style"],
                                },
                                "then": {"properties": {"explode": {"default": True}}},
                                "else": {"properties": {"explode": {"default": False}}},
                            },
                        },
                    }
                },
                "$ref": "#/$defs/specification-extensions",
                "unevaluatedProperties": False,
            },
            "parameter-or-reference": {
                "if": {"type": "object", "required": ["$ref"]},
                "then": {"$ref": "#/$defs/reference"},
                "else": {"$ref": "#/$defs/parameter"},
            },
            "request-body": {
                "$comment": "https://spec.openapis.org/oas/v3.1.0#request-body-object",
                "type": "object",
                "properties": {
                    "description": {"type": "string"},
                    "content": {"$ref": "#/$defs/content"},
                    "required": {"default": False, "type": "boolean"},
                },
                "required": ["content"],
                "$ref": "#/$defs/specification-extensions",
                "unevaluatedProperties": False,
            },
            "request-body-or-reference": {
                "if": {"type": "object", "required": ["$ref"]},
                "then": {"$ref": "#/$defs/reference"},
                "else": {"$ref": "#/$defs/request-body"},
            },
            "content": {
                "$comment": "https://spec.openapis.org/oas/v3.1.0#fixed-fields-10",
                "type": "object",
                "additionalProperties": {"$ref": "#/$defs/media-type"},
                "propertyNames": {"format": "media-range"},
            },
            "media-type": {
                "$comment": "https://spec.openapis.org/oas/v3.1.0#media-type-object",
                "type": "object",
                "properties": {
                    "schema": {"$dynamicRef": "#meta"},
                    "encoding": {
                        "type": "object",
                        "additionalProperties": {"$ref": "#/$defs/encoding"},
                    },
                },
                "allOf": [
                    {"$ref": "#/$defs/specification-extensions"},
                    {"$ref": "#/$defs/examples"},
                ],
                "unevaluatedProperties": False,
            },
            "encoding": {
                "$comment": "https://spec.openapis.org/oas/v3.1.0#encoding-object",
                "type": "object",
                "properties": {
                    "contentType": {"type": "string", "format": "media-range"},
                    "headers": {
                        "type": "object",
                        "additionalProperties": {"$ref": "#/$defs/header-or-reference"},
                    },
                    "style": {
                        "default": "form",
                        "enum": [
                            "form",
                            "spaceDelimited",
                            "pipeDelimited",
                            "deepObject",
                        ],
                    },
                    "explode": {"type": "boolean"},
                    "allowReserved": {"default": False, "type": "boolean"},
                },
                "allOf": [
                    {"$ref": "#/$defs/specification-extensions"},
                    {"$ref": "#/$defs/encoding/$defs/explode-default"},
                ],
                "unevaluatedProperties": False,
                "$defs": {
                    "explode-default": {
                        "if": {
                            "properties": {"style": {"const": "form"}},
                            "required": ["style"],
                        },
                        "then": {"properties": {"explode": {"default": True}}},
                        "else": {"properties": {"explode": {"default": False}}},
                    }
                },
            },
            "responses": {
                "$comment": "https://spec.openapis.org/oas/v3.1.0#responses-object",
                "type": "object",
                "properties": {"default": {"$ref": "#/$defs/response-or-reference"}},
                "patternProperties": {
                    "^[1-5](?:[0-9]{2}|XX)$": {"$ref": "#/$defs/response-or-reference"}
                },
                "minProperties": 1,
                "$ref": "#/$defs/specification-extensions",
                "unevaluatedProperties": False,
                "if": {
                    "$comment": "either default, or at least one response code property must exist",
                    "patternProperties": {"^[1-5](?:[0-9]{2}|XX)$": False},
                },
                "then": {"required": ["default"]},
            },
            "response": {
                "$comment": "https://spec.openapis.org/oas/v3.1.0#response-object",
                "type": "object",
                "properties": {
                    "description": {"type": "string"},
                    "headers": {
                        "type": "object",
                        "additionalProperties": {"$ref": "#/$defs/header-or-reference"},
                    },
                    "content": {"$ref": "#/$defs/content"},
                    "links": {
                        "type": "object",
                        "additionalProperties": {"$ref": "#/$defs/link-or-reference"},
                    },
                },
                "required": ["description"],
                "$ref": "#/$defs/specification-extensions",
                "unevaluatedProperties": False,
            },
            "response-or-reference": {
                "if": {"type": "object", "required": ["$ref"]},
                "then": {"$ref": "#/$defs/reference"},
                "else": {"$ref": "#/$defs/response"},
            },
            "callbacks": {
                "$comment": "https://spec.openapis.org/oas/v3.1.0#callback-object",
                "type": "object",
                "$ref": "#/$defs/specification-extensions",
                "additionalProperties": {"$ref": "#/$defs/path-item-or-reference"},
            },
            "callbacks-or-reference": {
                "if": {"type": "object", "required": ["$ref"]},
                "then": {"$ref": "#/$defs/reference"},
                "else": {"$ref": "#/$defs/callbacks"},
            },
            "example": {
                "$comment": "https://spec.openapis.org/oas/v3.1.0#example-object",
                "type": "object",
                "properties": {
                    "summary": {"type": "string"},
                    "description": {"type": "string"},
                    "value": True,
                    "externalValue": {"type": "string", "format": "uri"},
                },
                "not": {"required": ["value", "externalValue"]},
                "$ref": "#/$defs/specification-extensions",
                "unevaluatedProperties": False,
            },
            "example-or-reference": {
                "if": {"type": "object", "required": ["$ref"]},
                "then": {"$ref": "#/$defs/reference"},
                "else": {"$ref": "#/$defs/example"},
            },
            "link": {
                "$comment": "https://spec.openapis.org/oas/v3.1.0#link-object",
                "type": "object",
                "properties": {
                    "operationRef": {"type": "string", "format": "uri-reference"},
                    "operationId": {"type": "string"},
                    "parameters": {"$ref": "#/$defs/map-of-strings"},
                    "requestBody": True,
                    "description": {"type": "string"},
                    "body": {"$ref": "#/$defs/server"},
                },
                "oneOf": [
                    {"required": ["operationRef"]},
                    {"required": ["operationId"]},
                ],
                "$ref": "#/$defs/specification-extensions",
                "unevaluatedProperties": False,
            },
            "link-or-reference": {
                "if": {"type": "object", "required": ["$ref"]},
                "then": {"$ref": "#/$defs/reference"},
                "else": {"$ref": "#/$defs/link"},
            },
            "header": {
                "$comment": "https://spec.openapis.org/oas/v3.1.0#header-object",
                "type": "object",
                "properties": {
                    "description": {"type": "string"},
                    "required": {"default": False, "type": "boolean"},
                    "deprecated": {"default": False, "type": "boolean"},
                    "schema": {"$dynamicRef": "#meta"},
                    "content": {
                        "$ref": "#/$defs/content",
                        "minProperties": 1,
                        "maxProperties": 1,
                    },
                },
                "oneOf": [{"required": ["schema"]}, {"required": ["content"]}],
                "dependentSchemas": {
                    "schema": {
                        "properties": {
                            "style": {"default": "simple", "const": "simple"},
                            "explode": {"default": False, "type": "boolean"},
                        },
                        "$ref": "#/$defs/examples",
                    }
                },
                "$ref": "#/$defs/specification-extensions",
                "unevaluatedProperties": False,
            },
            "header-or-reference": {
                "if": {"type": "object", "required": ["$ref"]},
                "then": {"$ref": "#/$defs/reference"},
                "else": {"$ref": "#/$defs/header"},
            },
            "tag": {
                "$comment": "https://spec.openapis.org/oas/v3.1.0#tag-object",
                "type": "object",
                "properties": {
                    "name": {"type": "string"},
                    "description": {"type": "string"},
                    "externalDocs": {"$ref": "#/$defs/external-documentation"},
                },
                "required": ["name"],
                "$ref": "#/$defs/specification-extensions",
                "unevaluatedProperties": False,
            },
            "reference": {
                "$comment": "https://spec.openapis.org/oas/v3.1.0#reference-object",
                "type": "object",
                "properties": {
                    "$ref": {"type": "string", "format": "uri-reference"},
                    "summary": {"type": "string"},
                    "description": {"type": "string"},
                },
                "unevaluatedProperties": False,
            },
            "schema": {
                "$comment": "https://spec.openapis.org/oas/v3.1.0#schema-object",
                "$dynamicAnchor": "meta",
                "type": ["object", "boolean"],
            },
            "security-scheme": {
                "$comment": "https://spec.openapis.org/oas/v3.1.0#security-scheme-object",
                "type": "object",
                "properties": {
                    "type": {
                        "enum": [
                            "apiKey",
                            "http",
                            "mutualTLS",
                            "oauth2",
                            "openIdConnect",
                        ]
                    },
                    "description": {"type": "string"},
                },
                "required": ["type"],
                "allOf": [
                    {"$ref": "#/$defs/specification-extensions"},
                    {"$ref": "#/$defs/security-scheme/$defs/type-apikey"},
                    {"$ref": "#/$defs/security-scheme/$defs/type-http"},
                    {"$ref": "#/$defs/security-scheme/$defs/type-http-bearer"},
                    {"$ref": "#/$defs/security-scheme/$defs/type-oauth2"},
                    {"$ref": "#/$defs/security-scheme/$defs/type-oidc"},
                ],
                "unevaluatedProperties": False,
                "$defs": {
                    "type-apikey": {
                        "if": {
                            "properties": {"type": {"const": "apiKey"}},
                            "required": ["type"],
                        },
                        "then": {
                            "properties": {
                                "name": {"type": "string"},
                                "in": {"enum": ["query", "header", "cookie"]},
                            },
                            "required": ["name", "in"],
                        },
                    },
                    "type-http": {
                        "if": {
                            "properties": {"type": {"const": "http"}},
                            "required": ["type"],
                        },
                        "then": {
                            "properties": {"scheme": {"type": "string"}},
                            "required": ["scheme"],
                        },
                    },
                    "type-http-bearer": {
                        "if": {
                            "properties": {
                                "type": {"const": "http"},
                                "scheme": {
                                    "type": "string",
                                    "pattern": "^[Bb][Ee][Aa][Rr][Ee][Rr]$",
                                },
                            },
                            "required": ["type", "scheme"],
                        },
                        "then": {"properties": {"bearerFormat": {"type": "string"}}},
                    },
                    "type-oauth2": {
                        "if": {
                            "properties": {"type": {"const": "oauth2"}},
                            "required": ["type"],
                        },
                        "then": {
                            "properties": {"flows": {"$ref": "#/$defs/oauth-flows"}},
                            "required": ["flows"],
                        },
                    },
                    "type-oidc": {
                        "if": {
                            "properties": {"type": {"const": "openIdConnect"}},
                            "required": ["type"],
                        },
                        "then": {
                            "properties": {
                                "openIdConnectUrl": {"type": "string", "format": "uri"}
                            },
                            "required": ["openIdConnectUrl"],
                        },
                    },
                },
            },
            "security-scheme-or-reference": {
                "if": {"type": "object", "required": ["$ref"]},
                "then": {"$ref": "#/$defs/reference"},
                "else": {"$ref": "#/$defs/security-scheme"},
            },
            "oauth-flows": {
                "type": "object",
                "properties": {
                    "implicit": {"$ref": "#/$defs/oauth-flows/$defs/implicit"},
                    "password": {"$ref": "#/$defs/oauth-flows/$defs/password"},
                    "clientCredentials": {
                        "$ref": "#/$defs/oauth-flows/$defs/client-credentials"
                    },
                    "authorizationCode": {
                        "$ref": "#/$defs/oauth-flows/$defs/authorization-code"
                    },
                },
                "$ref": "#/$defs/specification-extensions",
                "unevaluatedProperties": False,
                "$defs": {
                    "implicit": {
                        "type": "object",
                        "properties": {
                            "authorizationUrl": {"type": "string", "format": "uri"},
                            "refreshUrl": {"type": "string", "format": "uri"},
                            "scopes": {"$ref": "#/$defs/map-of-strings"},
                        },
                        "required": ["authorizationUrl", "scopes"],
                        "$ref": "#/$defs/specification-extensions",
                        "unevaluatedProperties": False,
                    },
                    "password": {
                        "type": "object",
                        "properties": {
                            "tokenUrl": {"type": "string", "format": "uri"},
                            "refreshUrl": {"type": "string", "format": "uri"},
                            "scopes": {"$ref": "#/$defs/map-of-strings"},
                        },
                        "required": ["tokenUrl", "scopes"],
                        "$ref": "#/$defs/specification-extensions",
                        "unevaluatedProperties": False,
                    },
                    "client-credentials": {
                        "type": "object",
                        "properties": {
                            "tokenUrl": {"type": "string", "format": "uri"},
                            "refreshUrl": {"type": "string", "format": "uri"},
                            "scopes": {"$ref": "#/$defs/map-of-strings"},
                        },
                        "required": ["tokenUrl", "scopes"],
                        "$ref": "#/$defs/specification-extensions",
                        "unevaluatedProperties": False,
                    },
                    "authorization-code": {
                        "type": "object",
                        "properties": {
                            "authorizationUrl": {"type": "string", "format": "uri"},
                            "tokenUrl": {"type": "string", "format": "uri"},
                            "refreshUrl": {"type": "string", "format": "uri"},
                            "scopes": {"$ref": "#/$defs/map-of-strings"},
                        },
                        "required": ["authorizationUrl", "tokenUrl", "scopes"],
                        "$ref": "#/$defs/specification-extensions",
                        "unevaluatedProperties": False,
                    },
                },
            },
            "security-requirement": {
                "$comment": "https://spec.openapis.org/oas/v3.1.0#security-requirement-object",
                "type": "object",
                "additionalProperties": {"type": "array", "items": {"type": "string"}},
            },
            "specification-extensions": {
                "$comment": "https://spec.openapis.org/oas/v3.1.0#specification-extensions",
                "patternProperties": {"^x-": True},
            },
            "examples": {
                "properties": {
                    "example": True,
                    "examples": {
                        "type": "object",
                        "additionalProperties": {
                            "$ref": "#/$defs/example-or-reference"
                        },
                    },
                }
            },
            "map-of-strings": {
                "type": "object",
                "additionalProperties": {"type": "string"},
            },
        },
    }

    def __init__(self):
        super().__init__(def_name="openApiSpecification")
        self._schema = self._SCHEMA



"""
The Link API allows data to reference external resources.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, final, override

from betty.ancestry.description import HasDescription
from betty.ancestry.media_type import HasMediaType
from betty.json.schema import String
from betty.locale.localizable.gettext import _, ngettext
from betty.locale.localizable.linked_data import dump_linked_data
from betty.locale.localizable.property import LocalizableProperty
from betty.locale.localizable.static.schema import StaticTranslationsSchema
from betty.model import Entity, EntityDefinition
from betty.model.association import BidirectionalToZeroOrOne
from betty.privacy import HasPrivacy, Privacy, merge_privacies
from betty.property import Optional

if TYPE_CHECKING:
    from betty.ancestry.has_links import HasLinks
    from betty.json.linked_data import JsonLdObject
    from betty.locale.localizable import Localizable, ResolvableLocalizable
    from betty.media_type import MediaType
    from betty.portable import PortableMapping
    from betty.project import Project


@final
@EntityDefinition(
    "link",
    label=_("Link"),
    label_plural=_("Links"),
    label_countable=ngettext("{count} link", "{count} links"),
    public_facing=False,
)
class Link(HasMediaType, HasDescription, HasPrivacy, Entity):
    """
    .. plugin:: entity:link.
    """

    _url = LocalizableProperty(label=_("URL"))
    _label = Optional(LocalizableProperty(label=_("Label")))

    relationship: str | None
    """
    The link's `IANA link relationship <https://www.iana.org/assignments/link-relations/link-relations.xhtml>`_.
    """

    owner = BidirectionalToZeroOrOne["Link", "HasLinks"](
        "betty.ancestry.has_links:HasLinks",
        "links",
        label=_("Owner"),
    )
    """
    The entity hat owns the link.
    """

    def __init__(
        self,
        url: ResolvableLocalizable,
        *,
        relationship: str | None = None,
        label: ResolvableLocalizable | None = None,
        description: ResolvableLocalizable | None = None,
        media_type: MediaType | None = None,
        owner: HasLinks | None = None,
        privacy: Privacy | None = None,
    ):
        super().__init__(
            media_type=media_type, description=description, privacy=privacy
        )
        self._url = url
        self._label = label
        self.relationship = relationship
        if owner is not None:
            self.owner = owner

    @property
    def url(self) -> Localizable:
        """
        The URL the link points to.
        """
        return self._url

    @url.setter
    def url(self, url: ResolvableLocalizable) -> None:
        self._url = url

    @override
    @property
    def label(self) -> Localizable:
        """
        The human-readable short link label.
        """
        return self.url if self._label is None else self._label

    @label.setter
    def label(self, label: ResolvableLocalizable | None) -> None:
        self._label = label

    @label.deleter
    def label(self) -> None:
        del self._label

    @property
    def has_label(self) -> bool:
        """
        Whether the link has an explicit label set.
        """
        return self._label is not None

    @override
    async def dump_linked_data(self, project: Project, /) -> PortableMapping:
        public_localizers = await project.public_localizers
        portable = await super().dump_linked_data(project)
        if self.public:
            portable["url"] = dump_linked_data(self.url, localizers=public_localizers)
            if self._label is not None:
                portable["label"] = dump_linked_data(
                    self._label, localizers=public_localizers
                )
            if self.relationship is not None:
                portable["relationship"] = self.relationship
        return portable

    @override
    @classmethod
    async def linked_data_schema(cls, project: Project, /) -> JsonLdObject:
        schema = await super().linked_data_schema(project)
        schema.add_property(
            "url",
            StaticTranslationsSchema(
                title="Label", description="The full URL to the other resource."
            ),
            False,
        )
        schema.add_property(
            "relationship",
            String(
                description="The relationship between this resource and the link target (https://en.wikipedia.org/wiki/Link_relation)."
            ),
            False,
        )
        schema.add_property(
            "label",
            StaticTranslationsSchema(
                title="Label", description="The human-readable link label."
            ),
            False,
        )
        return schema

    @override
    def _get_effective_privacy(self) -> Privacy:
        privacy = super()._get_effective_privacy()
        if isinstance(self.owner, HasPrivacy):
            return merge_privacies(privacy, self.owner)
        return privacy



"""
Jobs.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, override

from betty.ancestry.link import Link
from betty.job import Job
from betty.project.load.jobs import PopulateLink

if TYPE_CHECKING:
    from betty.job.scheduler import Scheduler
    from betty.model import Entity
    from betty.project import Project


class PopulateEntity(Job):
    """
    Populate an entity.
    """

    def __init__(self, entity: Entity, *, project: Project):
        super().__init__(
            self.id_for(entity),
            dependents={PopulateLink.id_for(entity)}
            if isinstance(entity, Link)
            else None,
        )
        self._project = project
        self._entity = entity

    @classmethod
    def id_for(cls, entity: Entity) -> str:
        """
        Get the job ID.
        """
        return f"wiki:populate:{entity.plugin().id}:{entity.id}"

    @override
    async def do(self, scheduler: Scheduler, /) -> None:
        from betty.extension.wiki import Wiki

        extensions = await self._project.extensions
        populator = await extensions[Wiki].populator
        await populator.populate(self._entity)



"""
Mapping properties.
"""

from __future__ import annotations

from collections.abc import Callable, Mapping, MutableMapping
from typing import TYPE_CHECKING, Any, override

from betty.property import Property

if TYPE_CHECKING:
    from collections.abc import Callable

    from ty_extensions import Intersection

    from betty.data import Data, DataDefinition
    from betty.data.aggregate.collection.mapping import MappingDefinition
    from betty.locale.localizable import ResolvableLocalizable


class MappingProperty[MutableMappingT: MutableMapping[Any, Any], ValueSetT](
    Property[MutableMappingT, ValueSetT]
):
    """
    A property that contains a :py:class:`collections.abc.MutableMapping`.
    """

    _data: MappingDefinition[MutableMappingT]

    def __init__(
        self,
        data: Intersection[DataDefinition[MutableMappingT], MappingDefinition]
        | Data[Intersection[DataDefinition[MutableMappingT], MappingDefinition]],
        *,
        label: ResolvableLocalizable | None = None,
        description: ResolvableLocalizable | None = None,
        omit_load: bool | None = None,
        omit_dump: Callable[[MutableMappingT], bool] | None = None,
        default: Callable[[], Mapping] = dict,
    ):
        super().__init__(
            data,
            label=label,
            description=description,
            omit_load=omit_load,
            omit_dump=omit_dump,
            default=self._new_default,
        )
        self._default_values = default

    def _new_default(self) -> MutableMappingT:
        new = self._data.new()
        new.update(self._default_values())
        return new

    @override
    def set(self, instance: Any, value: ValueSetT | MutableMappingT) -> MutableMappingT:
        data = self.get(instance)
        data.clear()
        data.update(self._resolver(value))
        return data



"""
Jobs for the Webpack extension.
"""

from __future__ import annotations

from asyncio import to_thread
from shutil import copytree
from typing import TYPE_CHECKING, override

from betty.job import Context, Job

if TYPE_CHECKING:
    from pathlib import Path

    from betty.extension.webpack.build import Builder
    from betty.job.scheduler import Scheduler


class _GenerateAssets(Job):
    def __init__(self, *, builder: Builder, www_directory: Path, cache_directory: Path):
        super().__init__("webpack:generate-assets", priority=True)
        self._builder = builder
        self._cache_directory = cache_directory
        self._www_directory = www_directory

    @override
    async def do(self, scheduler: Scheduler, /) -> None:
        context = scheduler.context
        build_directory_path = await self._generate_ensure_build_directory(context)
        context._webpack_build_directory_path = build_directory_path  # ty:ignore[unresolved-attribute]
        await self._copy_build_directory(build_directory_path, self._www_directory)

    async def _copy_build_directory(
        self, build_directory: Path, destination_directory: Path
    ) -> None:
        await to_thread(
            copytree, build_directory, destination_directory, dirs_exist_ok=True
        )

    async def _generate_ensure_build_directory(self, context: Context) -> Path:
        return await self._builder.build(
            self._cache_directory,
            context=context,
        )



"""
Provide plugin assertions.
"""

from collections.abc import Collection
from typing import Any

from betty.assertion import AssertionChain, assert_str
from betty.exception import HumanFacingException
from betty.locale.localizable.gettext import _
from betty.locale.localizable.markup import Paragraph, do_you_mean
from betty.plugin import PluginDefinition


def assert_plugin[PluginDefinitionT: PluginDefinition](
    available_plugins: Collection[PluginDefinitionT],
) -> AssertionChain[Any, PluginDefinitionT]:
    """
    Assert that a value is a plugin ID.
    """

    def _assert(
        value: Any,
    ) -> PluginDefinitionT:
        plugin_id = assert_str()(value)
        for plugin in available_plugins:
            if plugin.id == plugin_id:
                return plugin
        raise HumanFacingException(
            Paragraph(
                _('Unknown plugin "{plugin_id}".').format(plugin_id=plugin_id),
                do_you_mean(*(f'"{plugin.id}"' for plugin in available_plugins)),
            )
        ) from None

    return AssertionChain(_assert)



"""
User sessions that do nothing.
"""

import logging
from collections.abc import AsyncIterator
from contextlib import asynccontextmanager
from typing import final, overload, override

from betty.assertion import Assertion
from betty.locale.localizable import ResolvableLocalizable
from betty.progress import Progress
from betty.progress.no_op import NoOpProgress
from betty.typing import Void, VoidType
from betty.user import User, UserTimeoutError, Verbosity


@final
class NoOpUser(User):
    """
    A user session that does nothing.
    """

    verbosity = Verbosity.DEFAULT

    @override
    async def set_verbosity(self, verbosity: Verbosity, /) -> None:
        self.verbosity = verbosity

    @override
    async def message_exception(self) -> None:
        pass

    @override
    async def message_error(self, message: ResolvableLocalizable, /) -> None:
        pass

    @override
    async def message_warning(self, message: ResolvableLocalizable, /) -> None:
        pass

    @override
    async def message_information(self, message: ResolvableLocalizable, /) -> None:
        pass

    @override
    async def message_information_details(
        self, message: ResolvableLocalizable, /
    ) -> None:
        pass

    @override
    async def message_debug(self, message: ResolvableLocalizable, /) -> None:
        pass

    @override
    async def message_log(self, message: logging.LogRecord, /) -> None:
        pass

    @override
    @asynccontextmanager
    async def message_progress(
        self, message: ResolvableLocalizable, /
    ) -> AsyncIterator[Progress]:
        yield NoOpProgress()

    @override
    async def ask_confirmation(
        self, statement: ResolvableLocalizable, *, default: bool = False
    ) -> bool:
        raise UserTimeoutError

    @overload
    async def ask_input(
        self,
        question: ResolvableLocalizable,
        *,
        default: str | VoidType = Void,
    ) -> str:
        pass

    @overload
    async def ask_input[T](
        self,
        question: ResolvableLocalizable,
        *,
        assertion: Assertion[str, T],
        default: str | VoidType = Void,
    ) -> T:
        pass

    @override
    async def ask_input[T](
        self,
        question: ResolvableLocalizable,
        *,
        assertion: Assertion[str, T] | None = None,
        default: str | T | VoidType = Void,
    ) -> str | T:
        raise UserTimeoutError



"""
Contents for the demonstration site.
"""

from typing import Self, override

from betty.content import ContentDefinition
from betty.content.contents import Template, TemplateBuild
from betty.document import Document
from betty.project import Project
from betty.service.factory import Manufacturable
from betty.service.requirement.project import require_project


@ContentDefinition(
    "-demo-incomplete-translation-warning", label="Incomplete translation warning"
)
class _IncompleteTranslationWarning(Template, Manufacturable):
    @override
    @classmethod
    @require_project
    async def new(cls, project: Project, /) -> Self:
        return cls(jinja=await project.jinja)

    @override
    async def build_template(self, document: Document) -> TemplateBuild:
        return "component/demo/-demo-incomplete-translation-warning.html.j2"



"""
Data for the :py:class:`betty.extension.gramps.Gramps` extension.
"""

from __future__ import annotations

from collections.abc import Mapping, MutableMapping
from pathlib import Path
from typing import TYPE_CHECKING, final

from betty.collection.mapping import MutableResolvedMapping
from betty.collection.mapping.adapter import MutableResolvedMappingAdapter
from betty.data import Data, Sample
from betty.data.aggregate.collection.mapping import MappingDefinition
from betty.data.aggregate.collection.sequence import SequenceDefinition
from betty.data.aggregate.record.object import ObjectDefinition
from betty.data.str import StrDefinition
from betty.event_type import EventTypeDefinition, EventTypeManufacturer
from betty.exception import HumanFacingException
from betty.gramps.loader import (
    DEFAULT_EVENT_TYPE_MAPPING,
    DEFAULT_PLACE_TYPE_MAPPING,
    DEFAULT_ROLE_MAPPING,
)
from betty.locale.localizable.gettext import _
from betty.pathlib import FilePathDefinition
from betty.place_type import PlaceTypeDefinition, PlaceTypeManufacturer
from betty.plugin import Plugin, PluginDefinition
from betty.plugin.factory import PluginManufacturer, ResolvablePluginManufacturer
from betty.property import (
    Optional,
    Property,
)
from betty.property.collection.mapping import MappingProperty
from betty.property.collection.sequence import SequenceProperty
from betty.role import RoleDefinition, RoleManufacturer
from betty.sample import Size

if TYPE_CHECKING:
    from collections.abc import Iterable

    from betty.event_type import EventType
    from betty.locale.localizable import ResolvableLocalizable
    from betty.place_type import PlaceType
    from betty.role import Role


class _PluginMappingProperty[PluginDefinitionT: PluginDefinition, PluginT: Plugin](
    MappingProperty[
        MutableMapping[str, PluginManufacturer[PluginDefinitionT, PluginT]],
        Mapping[str, ResolvablePluginManufacturer[PluginDefinitionT, PluginT]],
    ]
):
    def __init__(
        self,
        manufacturer: type[PluginManufacturer[PluginDefinitionT, PluginT]],
        gramps_label: ResolvableLocalizable,
        default: Mapping[str, ResolvablePluginManufacturer[PluginDefinitionT, PluginT]],
    ):
        super().__init__(
            MappingDefinition(
                cls=MutableResolvedMapping,
                factory=lambda: MutableResolvedMappingAdapter(
                    {},
                    value_resolver=manufacturer.resolve,
                ),
                key=StrDefinition(label=gramps_label),
                value=manufacturer,
                label=manufacturer.type().type().label_plural,
            ),
            default=lambda: MutableResolvedMappingAdapter(
                dict(
                    zip(
                        default.keys(),
                        manufacturer.resolve_sequence(default.values()),
                        strict=False,
                    )
                ),
                value_resolver=manufacturer.resolve,
            ),
        )


@final
@ObjectDefinition(
    label=_("Family tree"),
    samples=[
        lambda: Sample(
            FamilyTree(name="my-gramps-family-tree"), label="Minimal", size=Size.MINIMAL
        )
    ],
)
class FamilyTree(Data):
    """
    A Gramps family tree.

    .. data:: betty.extension.gramps.data:FamilyTree
    """

    event_types = _PluginMappingProperty(
        EventTypeManufacturer, _("Gramps event type"), DEFAULT_EVENT_TYPE_MAPPING
    )
    """
    How to map event types.
    """

    file = Optional(Property(FilePathDefinition(), label=_("File")))
    """
    The path to a Gramps family tree file.
    """

    name = Optional(Property(StrDefinition(label=_("Name"))))
    """
    The family tree's name in Gramps.
    """

    place_types = _PluginMappingProperty(
        PlaceTypeManufacturer, _("Gramps place type"), DEFAULT_PLACE_TYPE_MAPPING
    )
    """
    How to map place types.
    """

    roles = _PluginMappingProperty(
        RoleManufacturer, _("Gramps role"), DEFAULT_ROLE_MAPPING
    )
    """
    How to map presence roles.
    """

    def __init__(
        self,
        file: Path | None = None,
        name: str | None = None,
        event_types: Mapping[
            str, ResolvablePluginManufacturer[EventTypeDefinition, EventType]
        ]
        | None = None,
        place_types: Mapping[
            str, ResolvablePluginManufacturer[PlaceTypeDefinition, PlaceType]
        ]
        | None = None,
        roles: Mapping[str, ResolvablePluginManufacturer[RoleDefinition, Role]]
        | None = None,
    ):
        super().__init__()
        self.file = file
        self.name = name
        self.source  # noqa: B018
        if event_types is not None:
            self.event_types.update(event_types)  # ty:ignore[no-matching-overload]
        if place_types is not None:
            self.place_types.update(place_types)  # ty:ignore[no-matching-overload]
        if roles is not None:
            self.roles.update(roles)  # ty:ignore[no-matching-overload]

    @property
    def source(self) -> Path | str:
        """
        The family tree's source.

        This is either the name of a family tree in Gramps, or the path to a Gramps family tree file.
        """
        if self.file is not None:
            return self.file
        if self.name is not None:
            return self.name
        raise HumanFacingException(
            _('Family tree configuration must either have a "file" or a "name"')
        )


@final
@ObjectDefinition(
    label=_("Gramps configuration"),
    samples=[
        lambda: Sample(GrampsConfiguration(), label="Minimal", size=Size.MINIMAL),
        lambda: Sample(
            GrampsConfiguration(executable=Path("gramps.exe")),
            label="A custom Gramps executable",
        ),
        lambda: Sample(
            GrampsConfiguration(family_trees=[FamilyTree(file=Path("./gramps.gpkg"))]),
            label="Load a family tree from a file",
        ),
        lambda: Sample(
            GrampsConfiguration(family_trees=[FamilyTree(name="my-family-tree")]),
            label="Load a family tree by its name directly from Gramps",
        ),
        lambda: Sample(
            GrampsConfiguration(
                family_trees=[
                    FamilyTree(
                        name="my-family-tree",
                        event_types={"GrampsEventType": "betty-event-type"},
                    ),
                ]
            ),
            label="Map a Gramps event type to a Betty event type",
        ),
        lambda: Sample(
            GrampsConfiguration(
                family_trees=[
                    FamilyTree(
                        name="my-family-tree",
                        place_types={"GrampsPlaceType": "betty-place-type"},
                    ),
                ]
            ),
            label="Map a Gramps place type to a Betty place type",
        ),
        lambda: Sample(
            GrampsConfiguration(
                family_trees=[
                    FamilyTree(
                        name="my-family-tree",
                        event_types={"GrampsRole": "betty-role"},
                    ),
                ]
            ),
            label="Map a Gramps role to a Betty role",
        ),
    ],
)
class GrampsConfiguration(Data):
    """
    Configuration for the :py:class:`betty.extension.gramps.Gramps` extension.

    .. data:: betty.extension.gramps.data:GrampsConfiguration
    """

    family_trees = SequenceProperty(
        SequenceDefinition(cls=list, value=FamilyTree, label=_("Family trees")),
        omit_load=True,
        omit_dump=lambda data: not len(data),
    )
    """
    The Gramps family trees to load.
    """

    executable = Optional(Property(FilePathDefinition()))
    """
    The path to a specific Gramps executable.

    Leave ``None`` to use Gramps from the PATH.
    """

    def __init__(
        self,
        *,
        family_trees: Iterable[FamilyTree] | None = None,
        executable: Path | None = None,
    ):
        super().__init__()
        if family_trees is not None:
            self.family_trees = family_trees
        self.executable = executable



"""
Fetch information from Wikipedia.
"""

from __future__ import annotations

import re
from typing import TYPE_CHECKING, cast

from betty.locale.localizable.static import StaticTranslations

if TYPE_CHECKING:
    from collections.abc import Sequence

    from betty.link import Link
    from betty.locale.localize import Localizer


class NotAPageError(ValueError):
    """
    Raised when a URL does not point to a Wikipedia page.
    """


_PAGE_URL_PATTERN = re.compile(r"^https?://([a-z]+)\.wikipedia\.org/wiki/([^/?#]+).*$")


def parse_page_url(url: str) -> tuple[str, str]:
    """
    Parse the URL for a Wikipedia page.

    :return: A 2-tuple with the page language and the page name.
    """
    match = _PAGE_URL_PATTERN.fullmatch(url)
    if match is None:
        raise NotAPageError
    return cast(tuple[str, str], match.groups())


def parse_page_link(link: Link, localizers: Sequence[Localizer]) -> tuple[str, str]:
    """
    Parse the URL for a link to a Wikipedia page.

    :return: A 2-tuple with the page language and the page name.
    """
    original_urls = set(
        StaticTranslations.resolve(link.url, localizers).translations.values()
    )
    if len(original_urls) > 1:
        # Skip links that already provide different localized URLs, as things would get too complex.
        raise NotAPageError
    original_url = next(iter(original_urls))
    return parse_page_url(original_url)



"""
Handle requirements on projects.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

from betty.locale.localizable.gettext import _
from betty.service.requirement import Requirement, UnmetRequirement

if TYPE_CHECKING:
    from betty.project import Project
    from betty.service.level import ServiceLevel


async def _require_project(services: ServiceLevel, target: str, /) -> Project:
    from betty.project import Project

    if isinstance(services, Project):
        return services
    raise UnmetRequirement(_("{target} requires a project.").format(target=target))


require_project = Requirement(_require_project)



"""
Test utilities for :py:mod:`betty.model`.
"""

from __future__ import annotations

from typing import final

import pytest

from betty.locale import DEFAULT_LOCALE
from betty.locale.localizable.static import CountableStaticTranslations
from betty.locale.localize import DEFAULT_LOCALIZER
from betty.model import Entity, EntityDefinition


class EntityTestBase[EntityT: Entity]:
    """
    A base class for testing :py:class:`betty.model.Entity` implementations.
    """

    @pytest.fixture
    def sut(self) -> type[EntityT]:
        """
        Provide the system(s) under test.
        """
        raise NotImplementedError

    async def test_label(self, sut: EntityT) -> None:
        """
        Tests :py:meth:`betty.model.Entity.label` implementations.
        """
        assert sut.label.localize(DEFAULT_LOCALIZER)


@final
@EntityDefinition(
    "dummy-one",
    label="Dummy (one)",
    label_plural="Dummies (one)",
    label_countable=CountableStaticTranslations(
        {
            DEFAULT_LOCALE: {
                "one": "{count} dummy (one)",
                "other": "{count} dummies (one)",
            }
        }
    ),
)
class DummyEntityOne(Entity):
    """
    A dummy entity.
    """


@final
@EntityDefinition(
    "dummy",
    label="Dummy (two)",
    label_plural="Dummies (two)",
    label_countable=CountableStaticTranslations(
        {
            DEFAULT_LOCALE: {
                "one": "{count} dummy (two)",
                "other": "{count} dummies (two)",
            }
        }
    ),
)
class DummyEntityTwo(Entity):
    """
    A dummy entity.
    """


@final
@EntityDefinition(
    "dummy",
    label="Dummy (three)",
    label_plural="Dummies (three)",
    label_countable=CountableStaticTranslations(
        {
            DEFAULT_LOCALE: {
                "one": "{count} dummy (three)",
                "other": "{count} dummies (three)",
            }
        }
    ),
)
class DummyEntityThree(Entity):
    """
    A dummy entity.
    """


@final
@EntityDefinition(
    "dummy",
    label="Dummy (four)",
    label_plural="Dummies (four)",
    label_countable=CountableStaticTranslations(
        {
            DEFAULT_LOCALE: {
                "one": "{count} dummy (four)",
                "other": "{count} dummies (four)",
            }
        }
    ),
)
class DummyEntityFour(Entity):
    """
    A dummy entity.
    """


@final
@EntityDefinition(
    "dummy-non-public-facing-one",
    label="Dummy non-public-facing (two)",
    label_plural="Dummies non-public-facing (two)",
    label_countable=CountableStaticTranslations(
        {
            DEFAULT_LOCALE: {
                "one": "{count} dummy non-public-facing (one)",
                "other": "{count} dummies non-public-facing (one)",
            }
        }
    ),
    public_facing=False,
)
class DummyNonPublicFacingEntityOne(Entity):
    """
    A dummy non-public-facing entity.
    """



"""
URL handling within HTML.
"""

from collections.abc import Iterable

from lxml.html import fragment_fromstring, tostring

from betty.media_type.media_types import HTML
from betty.url import UrlGenerator


def generate_urls(
    html: str, attribute_names: Iterable[str], *, url_generator: UrlGenerator
) -> str:
    """
    Convert attribute values to URLs.
    """
    generated = False
    fragment = fragment_fromstring(html, create_parent="div")
    attributes_xpath = " or ".join(
        f"boolean(@{attribute_name})" for attribute_name in attribute_names
    )
    for element in fragment.xpath(f"//*[{attributes_xpath}]"):
        for attr_name in element.keys():  # noqa: SIM118
            if attr_name in attribute_names:
                attr_value = element.get(attr_name)
                if url_generator.supports(attr_value):
                    element.set(
                        attr_name, url_generator.generate(attr_value, media_type=HTML)
                    )
                    generated = True

    if generated:
        return tostring(fragment).decode("utf-8")[5:-6]
    return html



"""
Life cycle management.
"""

from __future__ import annotations

from asyncio import gather
from typing import TYPE_CHECKING, Any, final, override

from betty.asyncio import resolve_await
from betty.life_cycle import (
    AlreadyBootstrapped,
    Bootstrapper,
    LifeCycle,
    NotYetBootstrapped,
    Shutdowner,
)

if TYPE_CHECKING:
    from collections.abc import Collection, MutableSequence


class ManagedLifeCycle(LifeCycle):
    """
    An object that can bootstrap and shut down, and manage resources upon these events.

    Subclasses and third party code can manage resources using the
    :py:attr:`~betty.life_cycle.manage.ManagedLifeCycle.life_cycle` property, which ensures everything is done in order.
    """

    def __init__(self, *args: Any, **kwargs: Any):
        self.__life_cycle = LifeCycleManager()
        super().__init__(*args, **kwargs)

    @final
    @property
    def life_cycle(self) -> LifeCycleManager:
        """
        The life cycle manager.
        """
        return self.__life_cycle

    @final
    @override
    async def bootstrap(self) -> None:
        await super().bootstrap()
        await self.life_cycle.bootstrap()

    @final
    @override
    async def shutdown(self, *, wait: bool = True) -> None:
        await super().shutdown(wait=wait)
        await self.__life_cycle.shutdown(wait=wait)


@final
class LifeCycleManager(LifeCycle):
    """
    A life cycle manager.
    """

    def __init__(self):
        super().__init__()
        self._bootstrappers: MutableSequence[Collection[Bootstrapper]] = []
        self._shutdowners: MutableSequence[Collection[Shutdowner]] = []

    @override
    async def bootstrap(self) -> None:
        await super().bootstrap()
        for bootstrappers in self._bootstrappers:
            await gather(
                *[resolve_await(bootstrapper()) for bootstrapper in bootstrappers]
            )

    @override
    async def shutdown(self, *, wait: bool = True) -> None:
        await super().shutdown(wait=wait)
        for shutdowners in self._shutdowners:
            await gather(
                *[resolve_await(shutdowner(wait=wait)) for shutdowner in shutdowners]
            )

    def attach(self, *life_cycles: LifeCycle) -> None:
        """
        Attach a batch of other life cycles to this one.

        The life cycles within the batch will be bootstrapped and shut down concurrently.
        """
        self.assert_not_shut_down()
        bootstrappers = []
        shutdowners = []
        for life_cycle in life_cycles:
            life_cycle.assert_not_shut_down()
            if self.bootstrapped:
                if not life_cycle.bootstrapped:
                    raise NotYetBootstrapped(
                        f"{life_cycle} was not bootstrapped yet, but {self} was already."
                    )
            else:
                if life_cycle.bootstrapped:
                    raise AlreadyBootstrapped(
                        f"{life_cycle} was bootstrapped already, but {self} was not yet."
                    )
                bootstrappers.append(life_cycle.bootstrap)
            shutdowners.append(life_cycle.shutdown)
        self._bootstrappers.append(bootstrappers)
        self._shutdowners.append(shutdowners)

    def on_bootstrap(self, *bootstrappers: Bootstrapper) -> None:
        """
        Add a batch of bootstrap callbacks.

        The callbacks within the batch will be invoked concurrently.
        """
        self.assert_not_bootstrapped()
        if not bootstrappers:
            return
        self._bootstrappers.append(bootstrappers)

    def on_shutdown(self, *shutdowners: Shutdowner) -> None:
        """
        Add a batch of shutdown callbacks.

        The callbacks within the batch will be invoked concurrently.
        """
        self.assert_not_shut_down()
        if not shutdowners:
            return
        self._shutdowners.append(shutdowners)

    def on(self, *callbacks: tuple[Bootstrapper, Shutdowner]) -> None:
        """
        Add a batch of callbacks.

        The callbacks within the batch will be invoked concurrently.
        """
        bootstrappers, shutdowners = list(zip(*callbacks, strict=False))
        self.on_bootstrap(*bootstrappers)
        self.on_shutdown(*shutdowners)



"""
Plugins that can declare dependencies on other plugins.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

from betty.plugin import ResolvableId, resolve_id
from betty.plugin.ordered import OrderedPluginDefinition, sort_ordered_plugin_graph

if TYPE_CHECKING:
    from collections.abc import Iterable, Set
    from graphlib import TopologicalSorter

    from betty.machine_name import MachineName, ResolvableMachineName
    from betty.service.plugin import PluginManager


class DependentPluginDefinition[BaseClsT = Any](OrderedPluginDefinition[BaseClsT]):
    """
    A definition of a plugin that can declare its dependency on other plugins.
    """

    def __init__(
        self,
        plugin_id: ResolvableMachineName,
        *,
        comes_before: Set[ResolvableId] | None = None,
        comes_after: Set[ResolvableId] | None = None,
        depends_on: Set[ResolvableId] | None = None,
        **kwargs: Any,
    ):
        super().__init__(
            plugin_id, comes_before=comes_before, comes_after=comes_after, **kwargs
        )
        self._depends_on = (
            set()
            if depends_on is None
            else {resolve_id(plugin) for plugin in depends_on}
        )
        self._comes_after.update(self._depends_on)

    @property
    def depends_on(self) -> Set[MachineName]:
        """
        The plugins this one depends on.

        All plugins will automatically be added to :py:meth:`betty.plugin.ordered.OrderedPluginDefinition.comes_after`.
        """
        return self._depends_on


async def expand_plugin_dependencies[
    DependentPluginDefinitionT: DependentPluginDefinition
](
    available_plugins: PluginManager[DependentPluginDefinitionT],
    plugins: Iterable[DependentPluginDefinitionT],
    /,
) -> Set[DependentPluginDefinitionT]:
    """
    Expand a collection of plugins to include their dependencies.
    """
    dependencies = set()
    for plugin in plugins:
        dependencies.add(plugin)
        dependencies.update(
            await expand_plugin_dependencies(
                available_plugins,
                [
                    await available_plugins[depends_on]
                    for depends_on in plugin.depends_on
                ],
            )
        )
    return dependencies


async def sort_dependent_plugin_graph[
    DependentPluginDefinitionT: DependentPluginDefinition
](
    available_plugins: PluginManager[DependentPluginDefinitionT],
    plugins: Iterable[DependentPluginDefinitionT],
    /,
) -> TopologicalSorter[MachineName]:
    """
    Sort a dependent plugin graph.
    """
    return await sort_ordered_plugin_graph(
        available_plugins, await expand_plugin_dependencies(available_plugins, plugins)
    )



"""
Adapters between Betty and Python mappings.
"""

from collections.abc import (
    Callable,
    Iterable,
    Iterator,
    Mapping,
    MutableMapping,
    Sequence,
)
from contextlib import suppress
from itertools import chain
from typing import Any, final, overload, override

from betty.collection.mapping import MutableResolvedMapping, ResolvedMapping
from betty.functools import passthrough
from betty.typing import Void


class _ResolvedMappingAdapter[KeyT, ResolvableKeyT, ValueT](
    ResolvedMapping[KeyT, ResolvableKeyT, ValueT]
):
    def __init__(
        self,
        upstream: Mapping[KeyT, ValueT],
        *,
        key_resolver: Callable[[KeyT | ResolvableKeyT], KeyT] = passthrough,
    ):
        self._upstream = upstream
        self._key_resolver = key_resolver

    @final
    @override
    def __getitem__(self, key: KeyT | ResolvableKeyT) -> ValueT:
        return self._upstream[self._key_resolver(key)]

    @overload
    def get[T](self, key: KeyT | ResolvableKeyT, default: T, /) -> ValueT | T:
        pass

    @overload
    def get(self, key: KeyT | ResolvableKeyT, default: None = None, /) -> ValueT | None:
        pass

    @final
    @override
    def get(self, key, default=None):
        return self._upstream.get(self._key_resolver(key), default)

    @final
    @override
    def __iter__(self) -> Iterator[KeyT]:
        return iter(self._upstream)

    @final
    @override
    def __len__(self) -> int:
        return len(self._upstream)

    @final
    @override
    def __contains__(self, key: Any) -> bool:
        with suppress(Exception):
            key = self._key_resolver(key)
        return key in self._upstream


@final
class ResolvedMappingAdapter[KeyT, ResolvableKeyT, ValueT](
    _ResolvedMappingAdapter[KeyT, ResolvableKeyT, ValueT]
):
    """
    Decorate another mapping to resolve any values before proxying them.
    """


@final
class MutableResolvedMappingAdapter[KeyT, ResolvableKeyT, ValueT, ResolvableValueT](
    _ResolvedMappingAdapter[KeyT, ResolvableKeyT, ValueT],
    MutableResolvedMapping[KeyT, ResolvableKeyT, ValueT, ResolvableValueT],
):
    """
    Decorate another mapping to resolve any values before proxying them.
    """

    _upstream: MutableMapping[KeyT, ValueT]

    def __init__(
        self,
        upstream: MutableMapping[KeyT, ValueT],
        *,
        key_resolver: Callable[[KeyT | ResolvableKeyT], KeyT] = passthrough,
        value_resolver: Callable[[ValueT | ResolvableValueT], ValueT] = passthrough,
    ):
        super().__init__(upstream, key_resolver=key_resolver)
        self._value_resolver = value_resolver

    def __setitem__(
        self, key: KeyT | ResolvableKeyT, value: ValueT | ResolvableValueT
    ) -> None:
        self._upstream[self._key_resolver(key)] = self._value_resolver(value)

    def __delitem__(self, key: KeyT | ResolvableKeyT) -> None:
        del self._upstream[self._key_resolver(key)]

    @overload
    def update(
        self,
        other: Mapping[KeyT | ResolvableKeyT, ValueT | ResolvableValueT]
        | Iterable[tuple[KeyT | ResolvableKeyT, ValueT | ResolvableValueT]],
        /,
        **kwargs: ValueT | ResolvableValueT,
    ) -> None:
        pass

    @overload
    def update(self, **kwargs: ValueT | ResolvableValueT) -> None:
        pass

    @override
    def update(self, other=None, **kwargs) -> None:
        items = kwargs.items()
        if isinstance(other, Mapping):
            items = chain(items, other.items())
        elif isinstance(other, Sequence):
            items = chain(items, other)
        self._upstream.update(
            {
                self._key_resolver(key): self._value_resolver(value)  # ty:ignore[invalid-argument-type]
                for key, value in items
            }
        )

    @overload
    def setdefault[T](
        self: MutableMapping[KeyT, T | None],
        key: KeyT | ResolvableKeyT,
        default: None = None,
        /,
    ) -> T | None:
        pass

    @overload
    def setdefault(
        self, key: KeyT | ResolvableKeyT, default: ValueT | ResolvableValueT, /
    ) -> ValueT:
        pass

    @override
    def setdefault(self, key, default=Void):
        return self._upstream.setdefault(
            self._key_resolver(key),
            None if default is Void else self._value_resolver(default),
        )  # ty:ignore[no-matching-overload]

    @overload
    def pop(self, key: KeyT | ResolvableKeyT, /) -> ValueT:
        pass

    @overload
    def pop(
        self, key: KeyT | ResolvableKeyT, /, default: ValueT | ResolvableValueT
    ) -> ValueT:
        pass

    @overload
    def pop[T](self, key: KeyT | ResolvableKeyT, /, default: T) -> ValueT | T:
        pass

    @override
    def pop(self, key, default=Void):
        key = self._key_resolver(key)
        if default is Void:
            return self._upstream.pop(key)
        return self._upstream.pop(key, default)

    @override
    def popitem(self) -> tuple[KeyT, ValueT]:
        return self._upstream.popitem()



"""
Record data types.
"""

from __future__ import annotations

from abc import abstractmethod
from typing import TYPE_CHECKING, Any, Generic, Self, TypeVar, final, override

from betty.assertion import OptionalField, assert_mapping
from betty.data import DataDefinition, OptionalDefinition, Sample, Samples
from betty.data.aggregate import AggregateDefinition
from betty.data.indicator.selector import Element
from betty.locale.localizable import resolve_localizable
from betty.portable import Portable, PortableData, PortablePorter, Porter

if TYPE_CHECKING:
    from collections.abc import Callable, Iterable, MutableSequence, Sequence

    from betty.data import Data
    from betty.locale.localizable import Localizable, ResolvableLocalizable
    from betty.portable import PortableMapping


@final
class FieldDefinition[ElementT: Element[Any] = Element[Any], DataClsT = Any]:
    """
    A record field definition.
    """

    def __init__(
        self,
        selector: ElementT,
        data: DataDefinition[DataClsT] | type[Data[DataDefinition[DataClsT]]],
        *,
        label: ResolvableLocalizable | None = None,
        description: ResolvableLocalizable | None = None,
        omit_load: bool | None = None,
        omit_dump: Callable[[DataClsT], bool] | None = None,
    ):
        self._selector = selector
        self._data = data if isinstance(data, DataDefinition) else data.data()
        self._label = None if label is None else resolve_localizable(label)
        self._description = (
            None if description is None else resolve_localizable(description)
        )
        self._omit_load = omit_load
        self._omit_dump = omit_dump

    @property
    def selector(self) -> Element:
        """
        The field selector.
        """
        return self._selector

    @property
    def data(self) -> DataDefinition[DataClsT]:
        """
        The field's data definition.
        """
        return self._data

    @property
    def label(self) -> Localizable | None:
        """
        The human-readable field label.
        """
        return self._label

    @property
    def description(self) -> Localizable | None:
        """
        The human-readable long field description.
        """
        return self._description

    @property
    def omit_load(self) -> bool:
        """
        Check if the field may be omitted from the parent when loading from portable data.
        """
        if self._omit_load is not None:
            return self._omit_load
        return isinstance(self._data, OptionalDefinition)

    def omit_dump(self, data: DataClsT, /) -> bool:
        """
        Check if the field may be omitted from the parent when dumping to portable data.
        """
        if data is None and isinstance(self._data, OptionalDefinition):
            return True
        if self._omit_dump is None:
            return False
        return self._omit_dump(data)


_PortableRecordElementT = TypeVar(
    "_PortableRecordElementT", bound=Element[str], default=Element[str], covariant=True
)


class PortableRecord(Portable, Generic[_PortableRecordElementT]):
    """
    A record object capable of dumping and loading itself to and from portable data.
    """

    @classmethod
    @abstractmethod
    def load_key(
        cls, portable: PortableData, key: _PortableRecordElementT, portable_key: str, /
    ) -> Self:
        """
        Create a new instance from portable data and a portable primary key.

        :raises betty.exception.HumanFacingException: Raised if the portable data is invalid.
        """

    @abstractmethod
    def dump_key(self, key: _PortableRecordElementT, /) -> tuple[str, PortableData]:
        """
        Dump the instance to portable data and a portable primary key.

        :raises betty.portable.error.NotPortable: Raised if any part of the data is not portable.
        """


class RecordPorter[DataClsT = Any, ElementT: Element[str] = Element[str]](
    Porter[DataClsT]
):
    """
    An object capable of dumping and loading record data to and from portable data.
    """

    @abstractmethod
    def load_key(
        self, portable: PortableData, key: ElementT, portable_key: str, /
    ) -> DataClsT:
        """
        Create a new data instance from portable data and a portable primary key.

        :raises betty.exception.HumanFacingException: Raised if the portable data is invalid.
        """

    @abstractmethod
    def dump_key(self, data: DataClsT, key: ElementT, /) -> tuple[str, PortableData]:
        """
        Dump the data to portable data and a portable primary key.
        """


@final
class PortableRecordPorter[
    PortableRecordT: PortableRecord,
    ElementT: Element[str] = Element[str],
](PortablePorter[PortableRecordT], RecordPorter[PortableRecordT, ElementT]):
    """
    Expose a portable record data type as a porter.
    """

    @override
    def load_key(
        self, portable: PortableData, key: ElementT, portable_key: str, /
    ) -> PortableRecordT:
        return self._cls.load_key(portable, key, portable_key)

    @override
    def dump_key(
        self, data: PortableRecordT, key: ElementT, /
    ) -> tuple[str, PortableData]:
        return data.dump_key(key)


@final
class MappingPorter[DataClsT = Any, ElementT: Element[str] = Element[str]](
    RecordPorter[DataClsT]
):
    """
    Load and dump a record from and to portable mappings.
    """

    def __init__(self, record: RecordDefinition[DataClsT, ElementT], /):
        self._record = record

    @override
    def load(self, portable: PortableData, /) -> DataClsT:
        from betty.assertion import RequiredField, assert_record

        return self._record.factory(
            **assert_record(
                *[
                    (OptionalField if field.omit_load else RequiredField)(
                        field.selector.element, field.data.porter.load
                    )
                    for field in self._record.fields
                ]
            )(portable)
        )

    @override
    def dump(self, data: DataClsT, /) -> PortableMapping:
        portable = {}
        for field in self._record.fields:
            field_data = field.selector.get(data)
            if not field.omit_dump(field_data):
                portable[field.selector.element] = field.data.porter.dump(field_data)
        return portable

    @override
    def load_key(
        self,
        portable: PortableData,
        key: ElementT,
        portable_key: str,
        /,
    ) -> DataClsT:  # ty:ignore[invalid-method-override]
        return self.load({**assert_mapping()(portable), key.element: portable_key})

    @override
    def dump_key(
        self,
        data: DataClsT,
        key: ElementT,
        /,
    ) -> tuple[str, PortableData]:  # ty:ignore[invalid-method-override]
        portable = self.dump(data)
        portable_key = portable.pop(key.element)
        assert isinstance(portable_key, str)
        return portable_key, portable


class RecordDefinition[DataClsT = Any, ElementT: Element[str] = Element[str]](
    AggregateDefinition[DataClsT, ElementT]
):
    """
    A record data definition.

    Records have explicitly defined fields.
    """

    _porter: RecordPorter[DataClsT] | None

    def __init__(
        self,
        /,
        cls: type[DataClsT] | None = None,
        *,
        label: ResolvableLocalizable,
        fields: Sequence[FieldDefinition[ElementT, Any]] | None = None,
        description: ResolvableLocalizable | None = None,
        samples: Iterable[Callable[[], Sample[DataClsT]] | Samples] | None = None,
        factory: Callable[..., DataClsT] | None = None,
        porter: RecordPorter[DataClsT] | None = None,
    ):
        super().__init__(
            cls=cls,
            label=label,
            description=description,
            samples=samples,
            porter=porter,
        )
        self._factory = factory
        self._fields: MutableSequence[FieldDefinition[ElementT, Any]] = (
            [] if fields is None else list(fields)
        )

    @property
    def factory(self) -> Callable[..., DataClsT]:
        """
        The factory to create new instances.

        The factory's arguments are kwargs whose names are this record's field names, and whose values are their fully
        typed values.
        """
        return self.cls if self._factory is None else self._factory

    @override
    @property
    def porter(self) -> RecordPorter[DataClsT]:
        if self._porter is None:
            if issubclass(self.cls, PortableRecord):
                self._porter = PortableRecordPorter(self.cls)
            else:
                self._porter = MappingPorter(
                    self,  # ty:ignore[invalid-argument-type]
                )
        return self._porter

    @property
    def fields(self) -> Sequence[FieldDefinition[ElementT, Any]]:
        """
        The definitions of the fields contained by this record.
        """
        return self._fields



"""
Errors for the portable data API.
"""

from __future__ import annotations

from typing import final

from betty.exception import HumanFacingException


@final
class NotPortable(HumanFacingException):
    """
    Raised when data is not portable.
    """



"""
Progress tracking that does nothing.
"""

from typing import override

from betty.progress import Progress


class NoOpProgress(Progress):
    """
    A progress tracker that does nothing.
    """

    @override
    async def add(self, add: int = 1, /) -> None:
        pass

    @override
    async def done(self, done: int = 1, /) -> None:
        pass



"""
Definitions that are human-facing and provide human-friendly information.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

from betty.locale.localizable import resolve_localizable

if TYPE_CHECKING:
    from betty.locale.localizable import (
        CountableLocalizable,
        Localizable,
        ResolvableLocalizable,
    )


class HumanFacingDefinition:
    """
    A definition that is human-facing and provides human-friendly information.
    """

    def __init__(
        self,
        *args: Any,
        label: ResolvableLocalizable,
        description: ResolvableLocalizable | None = None,
        **kwargs: Any,
    ):
        super().__init__(*args, **kwargs)
        self._label = resolve_localizable(label)
        self._description = (
            None if description is None else resolve_localizable(description)
        )

    @property
    def label(self) -> Localizable:
        """
        The human-readable short label (singular).
        """
        return self._label

    @property
    def description(self) -> Localizable | None:
        """
        The human-readable long description.
        """
        return self._description


class CountableHumanFacingDefinition(HumanFacingDefinition):
    """
    A definition that is human-facing and provides countable human-friendly information.
    """

    def __init__(
        self,
        *args: Any,
        label: ResolvableLocalizable,
        label_plural: ResolvableLocalizable,
        label_countable: CountableLocalizable,
        description: ResolvableLocalizable | None = None,
        **kwargs: Any,
    ):
        super().__init__(*args, label=label, description=description, **kwargs)
        self._label_plural = resolve_localizable(label_plural)
        self._label_countable = label_countable

    @property
    def label_plural(self) -> Localizable:
        """
        The human-readable short label (plural).
        """
        return self._label_plural

    @property
    def label_countable(self) -> CountableLocalizable:
        """
        The human-readable short label (countable).
        """
        return self._label_countable



"""
Provide the Cache API.
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Awaitable, Callable
from typing import TYPE_CHECKING, Self

from betty.typing import threadsafe

if TYPE_CHECKING:
    from contextlib import AbstractAsyncContextManager


class CacheItem[CacheItemValueT](ABC):
    """
    A cache item.
    """

    @property
    @abstractmethod
    def modified(self) -> int | float:
        """
        Get the time this cache item was last modified, in seconds.
        """

    @abstractmethod
    async def value(self) -> CacheItemValueT:
        """
        Get this cache item's value.
        """


type CacheItemValueSetter[_CacheItemValueT] = Callable[
    [_CacheItemValueT], Awaitable[None]
]


@threadsafe
class Cache[CacheItemValueT](ABC):
    """
    A cache.

    To test your own subclasses, use :py:class:`betty.test_utils.cache.CacheTestBase`.
    """

    @abstractmethod
    def with_scope(self, scope: str, /) -> Self:
        """
        Return a new nested cache with the given scope.
        """

    async def has(self, cache_item_id: str, /) -> bool:
        """
        Check if a cache item with the given ID exists.
        """
        return await self.get(cache_item_id) is not None

    @abstractmethod
    def hasset(
        self, cache_item_id: str, /
    ) -> AbstractAsyncContextManager[CacheItemValueSetter[CacheItemValueT] | None]:
        """
        Check if a cache item with the given ID exists, and if not, provide a setter to add or update it within the same atomic operation.
        """

    @abstractmethod
    async def get(self, cache_item_id: str, /) -> CacheItem[CacheItemValueT] | None:
        """
        Get the cache item with the given ID.
        """

    @abstractmethod
    async def set(
        self,
        cache_item_id: str,
        value: CacheItemValueT,
        *,
        modified: int | float | None = None,
    ) -> None:
        """
        Add or update a cache item.
        """

    @abstractmethod
    def getset(
        self, cache_item_id: str, /
    ) -> AbstractAsyncContextManager[
        CacheItemValueSetter[CacheItemValueT] | CacheItem[CacheItemValueT]
    ]:
        """
        Get the cache item with the given ID, or provide a setter to add it within the same atomic operation.
        """

    @abstractmethod
    async def delete(self, cache_item_id: str, /) -> None:
        """
        Delete the cache item with the given ID.

        This operation s unsafe and MAY cause other concurrent operations on this cache to fail.
        """

    @abstractmethod
    async def clear(self) -> None:
        """
        Clear all items from the cache.

        This operation s unsafe and MAY cause other concurrent operations on this cache to fail.
        """



"""
Configuration file management.
"""

from __future__ import annotations

from contextlib import chdir
from typing import TYPE_CHECKING

import aiofiles
from aiofiles.os import makedirs

from betty.assertion import AssertionChain, assert_file_path
from betty.data.indicator import Path as DataPath
from betty.exception import reraise_with_indicator
from betty.serde import SerializerDefinition, serializer_for
from betty.service.level import UNIVERSE

if TYPE_CHECKING:
    from pathlib import Path

    from betty.portable import PortableData


async def assert_load_file() -> AssertionChain[Path, PortableData]:
    """
    An assertion to load a dump from a file.
    """
    available_formats = {
        available_format: await UNIVERSE.factory.new(available_format.cls)
        async for available_format in UNIVERSE.plugins[SerializerDefinition]
    }

    def _assert(file_path: Path) -> PortableData:
        with (
            reraise_with_indicator(DataPath(file_path)),
            # Change the working directory to allow relative paths to be resolved
            # against the configuration file's directory path.
            chdir(file_path.parent),
        ):
            with open(file_path) as f:
                dump_data = f.read()
            file_format = available_formats[
                serializer_for(list(available_formats), file_path.suffix)
            ]
            return file_format.load(dump_data)

    return assert_file_path() | _assert


async def dump_file(portable: PortableData, file_path: Path, /) -> None:
    """
    Write a dump to a file.
    """
    serializer = await UNIVERSE.factory.new(
        serializer_for(
            [plugin async for plugin in UNIVERSE.plugins[SerializerDefinition]],
            file_path.suffix,
        ).cls
    )
    dump_data = serializer.dump(portable)
    await makedirs(file_path.parent, exist_ok=True)
    async with aiofiles.open(file_path, mode="w") as f:
        await f.write(dump_data)



"""
The portable data API.

Portable data can easily be persistent or transmitted across and between systems.
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Callable, MutableMapping, MutableSequence
from typing import Any, Self, final, override

type PortableData = (
    bool
    | int
    | float
    | str
    | None
    | MutableSequence[PortableData]
    | MutableMapping[str, PortableData]
)
"""
Portable data.

Data of this type is portable and can easily be persisted or transmitted.
"""


type PortableSequence[PortableDataT: PortableData] = MutableSequence[PortableDataT]
"""
A sequence of portable data.
"""


type PortableMapping[PortableDataT: PortableData] = MutableMapping[str, PortableDataT]
"""
A key-value mapping of portable data.

Keys are strings.
"""


class Portable[PortableDataT: PortableData](ABC):
    """
    A class that can be dumped to and loaded from portable data.
    """

    @classmethod
    @abstractmethod
    def load(cls, portable: PortableData, /) -> Self:
        """
        Create a new instance from portable data.

        :raises betty.exception.HumanFacingException: Raised if the portable data is invalid.
        """

    @abstractmethod
    def dump(self) -> PortableDataT:
        """
        Dump the instance to portable data.

        :raises betty.portable.error.NotPortable: Raised if any part of the data is not portable.
        """


class Porter[DataClsT = Any, PortableDataT: PortableData = PortableData](ABC):
    """
    An object capable of dumping and loading data to and from portable data.
    """

    @abstractmethod
    def load(self, portable: PortableData, /) -> DataClsT:
        """
        Load data from its portable form.
        """

    @abstractmethod
    def dump(self, data: DataClsT, /) -> PortableDataT:
        """
        Dump data to its portable form.
        """

    def copy(self, data: DataClsT) -> DataClsT:
        """
        Deep-copy data into a new instance.
        """
        return self.load(self.dump(data))


@final
class CallbackPorter[DataClsT, PortableDataT: PortableData = PortableData](
    Porter[DataClsT, PortableDataT]
):
    """
    Make data portable using a separate loader and dumper.
    """

    def __init__(
        self,
        loader: Callable[[PortableData], DataClsT],
        dumper: Callable[[DataClsT], PortableDataT],
        /,
    ):
        self._loader = loader
        self._dumper = dumper

    @override
    def load(self, portable: PortableData) -> DataClsT:
        return self._loader(portable)

    @override
    def dump(self, data: DataClsT) -> PortableDataT:
        return self._dumper(data)


class PortablePorter[PortableT: Portable, PortableDataT: PortableData = PortableData](
    Porter[PortableT, PortableDataT]
):
    """
    Expose a portable data type as a porter.
    """

    def __init__(self, cls: type[PortableT]):
        self._cls = cls

    @override
    def load(self, portable: PortableData) -> PortableT:
        return self._cls.load(portable)

    @override
    def dump(self, data: PortableT) -> PortableDataT:
        return data.dump()


@final
class OptionalPorter[PortableT: Portable, PortableDataT: PortableData = PortableData](
    Porter[PortableT | None, PortableDataT | None]
):
    """
    Add optional (``None``) support to another porter.
    """

    def __init__(self, upstream: Porter[PortableT, PortableDataT]):
        self._upstream = upstream

    @override
    def load(self, portable: PortableData) -> PortableT | None:
        if portable is None:
            return None
        return self._upstream.load(portable)

    @override
    def dump(self, data: PortableT | None) -> PortableDataT | None:
        if data is None:
            return None
        return self._upstream.dump(data)



"""
Collection data tpes.
"""

from abc import abstractmethod
from collections.abc import Collection


class MutableCollection[ValueT](Collection[ValueT]):
    """
    A mutable collection of values.
    """

    @abstractmethod
    def clear(self) -> None:
        """
        Remove all values from the collection.
        """



from __future__ import annotations  # noqa: D100

from typing import TYPE_CHECKING, Self, final, override

from betty.console.command import Command, CommandDefinition, CommandFunction
from betty.console.project import add_project_argument
from betty.job import Context
from betty.locale.localizable.gettext import _
from betty.service.factory import Manufacturable
from betty.service.requirement.app import require_app

if TYPE_CHECKING:
    import argparse

    from betty.app import App
    from betty.project import Project


@final
@CommandDefinition("generate", label=_("Generate a static site"))
class Generate(Manufacturable, Command):
    """
    .. plugin:: command:generate.
    """

    def __init__(self, app: App):
        self._app = app

    @override
    @classmethod
    @require_app
    async def new(cls, app: App, /) -> Self:
        return cls(app)

    @override
    async def configure(self, parser: argparse.ArgumentParser) -> CommandFunction:
        return await add_project_argument(parser, self._command_function, self._app)

    async def _command_function(self, project: Project) -> None:
        from betty.project import generate, load

        async with (
            project,
            project.app.user.message_progress(_("Generating site...")) as progress,
        ):
            # Add a phantom value to the progress so it can never jump to 100% before we are entirely done here.
            await progress.add()

            context = Context(progress=progress)
            await load.load(project, context=context)
            await generate.generate(project, context=context)

            await progress.done()



"""
Test utilities for :py:mod:`betty.service.level`.
"""

from typing import Any, Self, override

from betty.service.factory import DataManufacturable, Manufacturable
from betty.service.level import ServiceLevel
from betty.test_utils.data import DummyData


class DummyDataManufacturable(DataManufacturable[DummyData], Manufacturable):
    """
    A dummy :py:class:`betty.service.factory.DataManufacturable` implementation.
    """

    def __init__(self, *args: Any, data: DummyData | None = None, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self.data = DummyData() if data is None else data

    @override
    @classmethod
    def new_data_cls(cls) -> type[DummyData]:
        return DummyData

    @override
    @classmethod
    async def new(
        cls, services: ServiceLevel, data: DummyData | None = None, /
    ) -> Self:
        return cls(data=data)



"""
Requirements for services.
"""

from __future__ import annotations

from functools import partial, update_wrapper
from typing import TYPE_CHECKING, Any, Concatenate, final, overload

from betty.asyncio import resolve_await
from betty.exception import HumanFacingException
from betty.importlib import fully_qualified_name
from betty.service.level import ServiceLevel

if TYPE_CHECKING:
    from collections.abc import Awaitable, Callable


@final
class Requirement[RequirementT]:
    """
    A service level requirement.
    """

    def __init__(
        self, require: Callable[[ServiceLevel, str], Awaitable[RequirementT]], /
    ):

        self._require = require

    @overload
    def __call__(self, f: ServiceLevel, /) -> Awaitable[RequirementT]:
        pass

    @overload
    def __call__[**P, ReturnT](
        self,
        f: Callable[Concatenate[RequirementT, P], Awaitable[ReturnT] | ReturnT],
    ) -> Callable[Concatenate[ServiceLevel, P], Awaitable[ReturnT]]:
        pass

    @overload
    def __call__[**P, ReturnT, MagicT](
        self,
        f: Callable[Concatenate[MagicT, RequirementT, P], Awaitable[ReturnT] | ReturnT],
    ) -> Callable[Concatenate[MagicT, ServiceLevel, P], Awaitable[ReturnT]]:
        pass

    def __call__[**P, ReturnT](
        self, f
    ) -> CallableRequirement[RequirementT, P, ReturnT] | Awaitable[RequirementT]:
        """
        Decorate a callable with this requirement.
        """
        if isinstance(f, ServiceLevel):
            return self._require(f, "**UNKNOWN**")
        return CallableRequirement(self._require, f)


@final
class CallableRequirement[RequirementT, **P, ReturnT, MagicT = Any]:
    """
    A requirement that can be called or used as a descriptor.
    """

    def __init__(
        self,
        require: Callable[[ServiceLevel, str], Awaitable[RequirementT]],
        f: Callable[Concatenate[RequirementT, P], Awaitable[ReturnT] | ReturnT]
        | Callable[Concatenate[MagicT, RequirementT, P], Awaitable[ReturnT] | ReturnT],
    ):
        self._require = require
        update_wrapper(self, f)
        self._f = f

    def __get__(self, instance, owner):
        if instance is None:
            return self
        return partial(self, instance)

    @overload
    async def __call__(
        self, requirement: RequirementT, *args: P.args, **kwargs: P.kwargs
    ) -> ReturnT:
        pass

    @overload
    async def __call__(
        self,
        magic: MagicT,
        requirement: RequirementT,
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> ReturnT:
        pass

    async def __call__(self, *args, **kwargs) -> ReturnT:
        """
        Call the decorated callable.
        """
        index = 0 if isinstance(args[0], ServiceLevel) else 1
        return await resolve_await(
            self._f(
                *args[:index],
                await self._require(args[index], fully_qualified_name(self._f)),
                *args[index + 1 :],
                **kwargs,
            )
        )


@final
class UnmetRequirement(HumanFacingException):
    """
    Raised when a requirement is not met.
    """



"""
JSON schemas for the project API.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Self, final, override

from betty.json.schema import JsonSchemaReference, Schema
from betty.model import EntityDefinition
from betty.model.schema import ToManySchema
from betty.service.factory import Manufacturable
from betty.service.requirement.project import require_project
from betty.string import kebab_case_to_lower_camel_case

if TYPE_CHECKING:
    from pathlib import Path

    from betty.project import Project


@final
class ProjectSchema(Manufacturable, Schema):
    """
    A JSON Schema for a project.
    """

    @classmethod
    async def def_url(cls, project: Project, def_name: str) -> str:
        """
        Get the URL to a project's JSON Schema definition.
        """
        return f"{await cls.url(project)}#/$defs/{def_name}"

    @classmethod
    async def url(cls, project: Project) -> str:
        """
        Get the URL to a project's JSON Schema.
        """
        url_generator = await project.url_generator
        return url_generator.generate("betty-static:///schema.json", absolute=True)

    @classmethod
    def www_path(cls, project: Project) -> Path:
        """
        Get the path to the schema file in a site's public WWW directory.
        """
        return project.www_directory / "schema.json"

    @override
    @classmethod
    @require_project
    async def new(cls, project: Project, /) -> Self:
        schema = cls()
        schema._schema["$id"] = await cls.url(project)

        # Add entity schemas.
        async for entity_type in project.plugins[EntityDefinition]:
            entity_type_schema = await entity_type.cls.linked_data_schema(project)
            entity_type_schema.embed(schema)
            def_name = f"{kebab_case_to_lower_camel_case(entity_type.id)}EntityCollectionResponse"
            schema.defs[def_name] = {
                "type": "object",
                "properties": {
                    "collection": ToManySchema().embed(schema),
                },
            }

        # Add the HTTP error response.
        schema.defs["errorResponse"] = {
            "type": "object",
            "properties": {
                "$schema": JsonSchemaReference().embed(schema),
                "message": {
                    "type": "string",
                },
            },
            "required": [
                "$schema",
                "message",
            ],
            "additionalProperties": False,
        }

        schema._schema["anyOf"] = [
            {"$ref": f"#/$defs/{def_name}"} for def_name in schema.defs
        ]

        return schema



"""
Provide Betty's ancestry place types.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, final, override

from betty.definition.human_facing import CountableHumanFacingDefinition
from betty.locale.localizable.gettext import _, ngettext
from betty.plugin import Plugin, PluginDefinition, PluginTypeDefinition
from betty.plugin.factory import PluginManufacturer

if TYPE_CHECKING:
    import builtins


class PlaceType(Plugin["PlaceTypeDefinition"]):
    """
    Define a :py:class:`betty.ancestry.place.Place` type.
    """


@final
@PluginTypeDefinition(
    "place-type",
    label=_("Place type"),
    label_plural=_("Place types"),
    label_countable=ngettext("{count} place type", "{count} place types"),
)
class PlaceTypeDefinition(CountableHumanFacingDefinition, PluginDefinition[PlaceType]):
    """
    .. plugin_type:: place-type.
    """


@final
class PlaceTypeManufacturer(PluginManufacturer[PlaceTypeDefinition, PlaceType]):
    """
    The place type manufacturer.
    """

    @override
    @classmethod
    def type(cls) -> builtins.type[PlaceTypeDefinition]:
        return PlaceTypeDefinition



"""
Data types for people's presences at events.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, final, override

from betty.locale.localizable.gettext import _, ngettext
from betty.model import Entity, EntityDefinition
from betty.model.association import BidirectionalToOne, ToOneAssociate
from betty.plugin.schema import PluginIdSchema
from betty.privacy import HasPrivacy, Privacy, is_public, merge_secondary_privacies
from betty.role import RoleDefinition

if TYPE_CHECKING:
    from betty.ancestry.event import Event
    from betty.ancestry.person import Person
    from betty.json.linked_data import JsonLdObject
    from betty.locale.localizable import Localizable
    from betty.portable import PortableMapping
    from betty.project import Project
    from betty.role import Role


@final
@EntityDefinition(
    "presence",
    label=_("Presence"),
    label_plural=_("Presences"),
    label_countable=ngettext("{count} presence", "{count} presences"),
    description=_("A person's presence at an event."),
    public_facing=False,
)
class Presence(HasPrivacy, Entity):
    """
    .. plugin:: entity:presence.
    """

    person = BidirectionalToOne["Presence", "Person"](
        "betty.ancestry.person:Person",
        "presences",
        label=_("Person"),
    )
    """
    The person whose presence is described.
    """

    event = BidirectionalToOne["Presence", "Event"](
        "betty.ancestry.event:Event",
        "presences",
        label=_("Event"),
    )
    """
    The event the person was present at.
    """

    role: Role
    """
    The role the person performed at the event.
    """

    def __init__(
        self,
        person: ToOneAssociate[Person],
        role: Role,
        event: ToOneAssociate[Event],
        *,
        privacy: Privacy | None = None,
    ):
        super().__init__(None, privacy=privacy)
        self.person = person
        self.role = role
        self.event = event

    @override
    @property
    def label(self) -> Localizable:
        return _("Presence of {person} at {event}").format(
            person=self.person.label,
            event=self.event.label,
        )

    @override
    def _get_effective_privacy(self) -> Privacy:
        return merge_secondary_privacies(
            super()._get_effective_privacy(),
            self.person,
            self.event,
        )

    @override
    @classmethod
    async def linked_data_schema(cls, project: Project, /) -> JsonLdObject:
        schema = await super().linked_data_schema(project)
        schema.add_property(
            "role",
            PluginIdSchema(
                RoleDefinition.type(),
                [x async for x in project.plugins[RoleDefinition]],
            ),
            False,
        )
        return schema

    @override
    async def dump_linked_data(self, project: Project, /) -> PortableMapping:
        portable = await super().dump_linked_data(project)
        if is_public(self):
            portable["role"] = self.role.plugin().id
        return portable



"""
System directories for Betty to store data in.

These directories must only be used in production/live environments, and must therefore be injected into anything
that needs it, so during testing temporary, isolated directories can be used.
"""

from __future__ import annotations

from pathlib import Path

import platformdirs

from betty.about import IS_DEVELOPMENT, VERSION_MAJOR

_APPNAME = "betty"
_APPAUTHOR = "betty"

ROOT_DIRECTORY_PATH = Path(__file__).resolve().parents[1]
"""
The betty installation root directory path.

This is read-only.
"""

ASSETS_DIRECTORY_PATH = ROOT_DIRECTORY_PATH / "assets"
"""
The betty installation assets directory path.

This is read-only.
"""

CACHE_DIRECTORY_PATH = platformdirs.user_cache_path(_APPNAME, _APPAUTHOR, VERSION_MAJOR)
"""
The Betty instance's cache directory path on the local system.
"""

APP_CONFIG_DIRECTORY_PATH = platformdirs.user_config_path(
    _APPNAME, _APPAUTHOR, VERSION_MAJOR
)
"""
The Betty instance's :py:class:`betty.app.App` configuration path on the local system.
"""

if IS_DEVELOPMENT:
    DEV_OUTPUT_DIRECTORY_PATH = ROOT_DIRECTORY_PATH / "dev-output"
    """
    The output directory for artifacts generated by development tools.
    """



"""Provides Betty with SPDX data."""

from __future__ import annotations

from typing import TYPE_CHECKING, final

from betty.app import App
from betty.extension import Extension, ExtensionDefinition
from betty.license.licenses import SpdxLicenseBuilder
from betty.locale.localizable.gettext import _
from betty.service.requirement.extension import require_extension
from betty.service.requirement.project import require_project

if TYPE_CHECKING:
    from collections.abc import Iterable

    from betty.license import LicenseDefinition
    from betty.project import Project


@final
@ExtensionDefinition(
    "spdx",
    label=_("SPDX licenses"),
    description=_(
        "Provide license plugins from the SPDX License List (https://spdx.org/licenses/) "
    ),
)
class Spdx(Extension[App]):
    """
    .. plugin:: extension:spdx.
    """

    # Provide an initializer without arguments so the factory can call it.
    def __init__(self):
        super().__init__()


@require_project
async def discover_licenses(project: Project) -> Iterable[LicenseDefinition]:
    """
    Discover the SPDX licenses.
    """
    await require_extension(Spdx)(project)
    return [
        license
        async for license in SpdxLicenseBuilder(  # noqa: A001
            binary_file_cache=project.app.binary_file_cache.with_scope("spdx"),
            http_client=await project.app.http_client,
            user=project.app.user,
        ).build()
    ]



"""
Plugin discovery.
"""

from __future__ import annotations

from asyncio import gather
from collections.abc import Awaitable, Callable, Iterable

from betty.asyncio import resolve_await
from betty.plugin import Plugin, PluginDefinition, ResolvableDefinition
from betty.service.level import ServiceLevel

type ResolvableDiscovery[PluginDefinitionT: PluginDefinition = PluginDefinition] = (
    Callable[
        [ServiceLevel],
        Awaitable[Iterable["ResolvableDiscovery[PluginDefinitionT]"]]
        | Iterable["ResolvableDiscovery[PluginDefinitionT]"],
    ]
    | ResolvableDefinition[PluginDefinitionT]
)


async def discover[PluginDefinitionT: PluginDefinition](
    services: ServiceLevel, *discoveries: ResolvableDiscovery[PluginDefinitionT]
) -> Iterable[PluginDefinitionT]:
    """
    Discover plugins definitions.
    """
    return [
        plugin
        for plugins in await gather(
            *[_discover(discovery, services) for discovery in discoveries]
        )
        for plugin in plugins
    ]


async def _discover[PluginDefinitionT: PluginDefinition](
    discovery: ResolvableDiscovery[PluginDefinitionT], services: ServiceLevel
) -> Iterable[PluginDefinitionT]:
    from betty.service.requirement import UnmetRequirement

    try:
        if isinstance(discovery, PluginDefinition):
            return [discovery]  # ty:ignore[invalid-return-type]
        if isinstance(discovery, type) and issubclass(discovery, Plugin):
            return [discovery.plugin()]  # ty:ignore[invalid-argument-type,invalid-return-type]
        return await discover(services, *await resolve_await(discovery(services)))
    except UnmetRequirement:
        return ()



"""
Betty's default job scheduler.
"""

from __future__ import annotations

from collections import defaultdict
from collections.abc import Iterable
from contextlib import asynccontextmanager
from graphlib import CycleError, TopologicalSorter
from typing import TYPE_CHECKING, cast, final, override

from betty.classtools import Singleton
from betty.concurrent import AsynchronizedLock, backoff
from betty.job import Context, Job
from betty.job.scheduler import (
    Cancelled,
    Completed,
    CyclicDependencyError,
    DuplicateJobError,
    Released,
    ScheduledJobBatch,
    Scheduler,
    UnknownJobError,
)
from betty.typing import threadsafe

if TYPE_CHECKING:
    from collections.abc import (
        AsyncIterator,
        Awaitable,
        Callable,
        MutableMapping,
        MutableSequence,
        Sequence,
    )

    from betty.user import User


class _ScheduledJobBatch:
    def __init__(
        self,
        scheduler: Scheduler,
        user: User,
        done: Callable[[Sequence[str]], Awaitable[None]],
        jobs: Sequence[Job],
        /,
    ):
        self._scheduler = scheduler
        self._user = user
        self._done = done
        self._jobs = jobs

    async def __call__(self) -> None:
        try:
            for job in self._jobs:
                await self._user.message_debug(f'Doing job "{job.id}"...')
                await job.do(self._scheduler)
        except BaseException as reason:
            await self._scheduler.cancel(reason)
            return
        await self._done([job.id for job in self._jobs])


@final
class _UnknownJob(Singleton):
    pass


@final
@threadsafe
class DefaultScheduler(Scheduler):
    """
    Betty's default job scheduler.
    """

    def __init__(
        self,
        *,
        user: User,
        context: Context | None = None,
    ):
        super().__init__(context)
        self._user = user
        self._lock = AsynchronizedLock.new_threadsafe()
        self._released = False
        self._cancelled = False
        self._cancelled_reason: BaseException | None = None
        self._completed = False
        self._jobs: MutableMapping[str, Job | _UnknownJob] = defaultdict(_UnknownJob)
        self._job_dependencies: MutableMapping[str, set[str]] = defaultdict(set)
        self._scheduled_jobs: set[str] = set()
        self._releasable_jobs_queue: MutableSequence[str] = []
        self._releasable_jobs_sorter: TopologicalSorter[str] | None = None
        self._released_jobs: set[str] = set()
        self._done_jobs: set[str] = set()

    def _is_open(self) -> bool:
        return not self._cancelled and not self._completed

    def _assert_open(self) -> None:
        self._assert_not_cancelled()
        self._assert_not_completed()

    def _assert_not_cancelled(self) -> None:
        if self._cancelled:
            if isinstance(self._cancelled_reason, Cancelled):
                raise self._cancelled_reason
            raise Cancelled from self._cancelled_reason

    def _assert_not_completed(self) -> None:
        if self._completed:
            raise Completed

    @override
    async def add(self, *jobs: Job) -> None:
        async with self._lock:
            self._assert_open()
            self._releasable_jobs_sorter = None
            for job in jobs:
                if not isinstance(self._jobs[job.id], _UnknownJob):
                    raise DuplicateJobError(job.id)
                if job.dependents:
                    if self._released:
                        raise Released
                    for dependent in job.dependents:
                        self._job_dependencies[dependent].add(job.id)
                self._jobs[job.id] = job
                self._scheduled_jobs.add(job.id)
                self._job_dependencies[job.id].update(
                    job.dependencies - self._done_jobs
                )
            await self._context.progress.add(len(jobs))

    @override
    async def release(self) -> None:
        async with self._lock:
            self._released = True

    def _update_releasable_jobs(self) -> None:
        if self._releasable_jobs_sorter is None:
            self._releasable_jobs_sorter = TopologicalSorter(self._job_dependencies)
            try:
                self._releasable_jobs_sorter.prepare()
            except CycleError as error:
                raise CyclicDependencyError(error.args[1]) from None

        possibly_newly_releasable_job_ids = self._releasable_jobs_sorter.get_ready()
        if possibly_newly_releasable_job_ids:
            newly_releasable_job_ids = set()
            unknown_job_id: str | None = None
            for possibly_newly_releasable_job_id in possibly_newly_releasable_job_ids:
                # A job may have been marked done since the last update.
                if possibly_newly_releasable_job_id in self._done_jobs:
                    self._releasable_jobs_sorter.done(possibly_newly_releasable_job_id)
                # Ignore jobs that have been issued already.
                elif possibly_newly_releasable_job_id in self._released_jobs:
                    continue
                else:
                    possibly_newly_releasable_job = self._jobs[
                        possibly_newly_releasable_job_id
                    ]
                    if isinstance(possibly_newly_releasable_job, _UnknownJob):
                        unknown_job_id = possibly_newly_releasable_job_id
                        continue
                    self._scheduled_jobs.discard(possibly_newly_releasable_job_id)
                    newly_releasable_job_ids.add(possibly_newly_releasable_job_id)
            self._releasable_jobs_queue = [
                job.id
                for job in sorted(
                    cast(
                        Iterable[Job],
                        (
                            self._jobs[job_id]
                            for job_id in {
                                *self._releasable_jobs_queue,
                                *newly_releasable_job_ids,
                            }
                            if not isinstance(self._jobs[job_id], _UnknownJob)
                        ),
                    ),
                    key=lambda job: job.priority,
                    reverse=True,
                )
            ]
            if (
                not self._releasable_jobs_queue
                and not self._released_jobs
                and unknown_job_id is not None
            ):
                raise UnknownJobError(unknown_job_id)

    @override  # noqa: RET503
    async def get(self) -> ScheduledJobBatch:
        async with self._cancel_on_exception():
            async for _ in backoff():
                async with self._lock:
                    batch = await self._get()
                    if batch is not None:
                        return batch
        raise RuntimeError("This never happpens.")

    async def _get(self) -> ScheduledJobBatch | None:
        self._assert_open()
        if not self._released:
            return None
        self._update_releasable_jobs()

        jobs: MutableSequence[Job] = []
        index = 0
        releasable_job_count = len(self._releasable_jobs_queue)
        while len(jobs) <= 9 and index < releasable_job_count:
            releasable_job = self._jobs[self._releasable_jobs_queue[index]]
            assert isinstance(releasable_job, Job)
            if jobs and (jobs[0].priority or releasable_job.priority):
                break
            self._releasable_jobs_queue.pop(index)
            releasable_job_count -= 1
            self._released_jobs.add(releasable_job.id)
            jobs.append(releasable_job)
        if jobs:
            return _ScheduledJobBatch(self, self._user, self._done, jobs)
        return None

    async def _done(self, job_ids: Sequence[str], /) -> None:
        async with self._cancel_on_exception():
            async with self._lock:
                for job_id in job_ids:
                    if self._releasable_jobs_sorter is not None:
                        self._releasable_jobs_sorter.done(job_id)
                    self._released_jobs.remove(job_id)
                    self._done_jobs.add(job_id)
                    self._job_dependencies.pop(job_id)
                    for job_dependencies in self._job_dependencies.values():
                        job_dependencies.discard(job_id)
            await self._context.progress.done(len(job_ids))

    @asynccontextmanager
    async def _cancel_on_exception(self) -> AsyncIterator[None]:
        try:
            yield
        except BaseException as exception:
            await self.cancel(exception)
            raise

    @override
    async def cancel(self, reason: BaseException | None = None, /) -> None:
        async with self._lock:
            if not self._is_open():
                return
            self._assert_open()
            self._cancelled = True
            self._cancelled_reason = reason

    @override
    async def complete(self) -> None:
        async with self._cancel_on_exception():
            async for _ in backoff():
                async with self._lock:
                    if self._completed:
                        return
                    self._assert_not_cancelled()
                    if (
                        not self._scheduled_jobs
                        and not self._releasable_jobs_queue
                        and not self._released_jobs
                    ):
                        self._completed = True
                        return



"""
Data for applications.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Final, final

from betty.data import Data, Sample
from betty.data.aggregate.record.object import ObjectDefinition
from betty.dirs import APP_CONFIG_DIRECTORY_PATH
from betty.locale import DEFAULT_LOCALE, ResolvableLocale, resolve_locale
from betty.locale.data import LocaleProperty
from betty.locale.localizable.gettext import _
from betty.property import Optional
from betty.sample import Size

if TYPE_CHECKING:
    from pathlib import Path


@final
@ObjectDefinition(
    label=_("Application configuration"),
    samples=[
        lambda: Sample(AppConfiguration(), label="Minimal", size=Size.MINIMAL),
        lambda: Sample(
            AppConfiguration(locale=DEFAULT_LOCALE), label="Full", size=Size.FULL
        ),
    ],
)
class AppConfiguration(Data):
    """
    Configuration for :py:class:`betty.app.App`.

    .. data:: betty.app.data:AppConfiguration
    """

    FILE: Final[Path] = APP_CONFIG_DIRECTORY_PATH / "app.json"

    locale = Optional(LocaleProperty())
    """
    The application locale.
    """

    def __init__(self, *, locale: ResolvableLocale | None = None):
        self.locale = None if locale is None else resolve_locale(locale)



"""
Provide tools to integrate extensions with `npm <https://www.npmjs.com/>`_.

This module is internal.
"""

from __future__ import annotations

import sys
from typing import TYPE_CHECKING

from betty import subprocess
from betty.exception import HumanFacingException
from betty.locale.localizable.gettext import _

if TYPE_CHECKING:
    from asyncio import subprocess as aiosubprocess
    from collections.abc import Sequence
    from pathlib import Path

    from betty.user import User


class NpmUnavailable(HumanFacingException, RuntimeError):
    """
    An error raised when npm is unavailable.
    """

    def __init__(self):
        super().__init__(
            _(
                "npm (https://www.npmjs.com/) must be available for features that require Node.js packages to be installed. Ensure that the `npm` executable is available in your `PATH`."
            )
        )


async def npm(
    arguments: Sequence[str], cwd: Path | None = None, *, user: User
) -> aiosubprocess.Process:
    """
    Run an npm command.
    """
    try:
        return await subprocess.run_process(
            ["npm", *arguments],
            cwd=cwd,
            # Use a shell on Windows so subprocess can find the executables it needs (see
            # https://bugs.python.org/issue17023).
            shell=sys.platform.startswith("win32"),
            user=user,
        )
    except FileNotFoundError:
        raise NpmUnavailable() from None



"""
Plugins that can declare their order.
"""

from __future__ import annotations

from collections import defaultdict
from graphlib import TopologicalSorter
from typing import TYPE_CHECKING, Any

from betty.machine_name import MachineName, ResolvableMachineName
from betty.plugin import PluginDefinition, ResolvableId, resolve_id

if TYPE_CHECKING:
    from collections.abc import Iterable, Iterator, Mapping, Set

    from betty.service.plugin import PluginManager


class OrderedPluginDefinition[BaseClsT](PluginDefinition[BaseClsT]):
    """
    A definition of plugin that can declare its order with respect to other plugins.
    """

    def __init__(
        self,
        plugin_id: ResolvableMachineName,
        *,
        comes_before: Set[ResolvableId] | None = None,
        comes_after: Set[ResolvableId] | None = None,
        **kwargs: Any,
    ):
        super().__init__(plugin_id, **kwargs)
        self._comes_before = (
            set()
            if comes_before is None
            else {resolve_id(plugin) for plugin in comes_before}
        )
        self._comes_after = (
            set()
            if comes_after is None
            else {resolve_id(plugin) for plugin in comes_after}
        )

    @property
    def comes_before(self) -> Set[MachineName]:
        """
        Get the plugins that this plugin comes before.

        The returned plugins come after this plugin.
        """
        return self._comes_before

    @property
    def comes_after(self) -> Set[MachineName]:
        """
        Get the plugins that this plugin comes after.

        The returned plugins come before this plugin.
        """
        return self._comes_after


async def sort_ordered_plugin_graph[OrderedPluginDefinitionT: OrderedPluginDefinition](
    available_plugins: PluginManager[OrderedPluginDefinitionT],
    plugins: Iterable[OrderedPluginDefinitionT],
    /,
) -> TopologicalSorter[MachineName]:
    """
    Build a graph of the given plugins.
    """
    sorter = TopologicalSorter[MachineName]()
    plugins = sorted(plugins, key=lambda plugin: plugin.id)
    for plugin in plugins:
        sorter.add(plugin.id)
        for before_identifier in map(resolve_id, plugin.comes_before):
            before = await available_plugins[before_identifier]
            if before in plugins:
                sorter.add(before.id, plugin.id)
        for after_identifier in map(resolve_id, plugin.comes_after):
            after = await available_plugins[after_identifier]
            if after in plugins:
                sorter.add(plugin.id, after.id)
    return sorter


async def get_comes_before[OrderedPluginDefinitionT: OrderedPluginDefinition](
    available_plugins: PluginManager[OrderedPluginDefinitionT],
    origin: OrderedPluginDefinitionT,
    /,
) -> Set[OrderedPluginDefinitionT]:
    """
    Get all other plugins the given plugin comes before.
    """
    graph = defaultdict(set)
    async for plugin in available_plugins:
        for comes_before_id in plugin.comes_before:
            comes_before = await available_plugins[comes_before_id]
            graph[plugin].add(comes_before)
        for comes_after_id in plugin.comes_after:
            comes_after = await available_plugins[comes_after_id]
            graph[comes_after].add(plugin)
    return set(_collect_plugin_graph(graph, origin))


async def get_comes_after[OrderedPluginDefinitionT: OrderedPluginDefinition](
    available_plugins: PluginManager[OrderedPluginDefinitionT],
    origin: OrderedPluginDefinitionT,
    /,
) -> Set[OrderedPluginDefinitionT]:
    """
    Get all other plugins the given plugin comes after.
    """
    graph = defaultdict(set)
    async for plugin in available_plugins:
        for comes_after_id in plugin.comes_after:
            comes_after = await available_plugins[comes_after_id]
            graph[plugin].add(comes_after)
        for comes_before_id in plugin.comes_before:
            comes_before = await available_plugins[comes_before_id]
            graph[comes_before].add(plugin)
    return set(_collect_plugin_graph(graph, origin))


def _collect_plugin_graph[PluginDefinitionT: PluginDefinition](
    graph: Mapping[PluginDefinitionT, Set[PluginDefinitionT]],
    origin: PluginDefinitionT,
) -> Iterator[PluginDefinitionT]:
    yield from graph[origin]
    for target in graph[origin]:
        yield from _collect_plugin_graph(graph, target)



"""
Adapters between keyed collections and Python data types.
"""

from collections.abc import Callable, Iterable, Iterator, Mapping
from contextlib import suppress
from typing import Any, final, override

from betty.collection.keyed import KeyedCollection, MutableKeyedCollection
from betty.functools import passthrough


class _KeyedCollectionAdapter[KeyT, ResolvableKeyT, ValueT](
    KeyedCollection[KeyT, ResolvableKeyT, ValueT]
):
    def __init__(
        self,
        values: Mapping[KeyT, ValueT] | None = None,
        *,
        key_resolver: Callable[[ResolvableKeyT | KeyT], KeyT] = passthrough,
    ):
        self._values = {} if values is None else dict(values)
        self._key_resolver = key_resolver

    @override
    def __len__(self) -> int:
        return self._values.__len__()

    @override
    def __iter__(self) -> Iterator[ValueT]:
        yield from self._values.values()

    @override
    def __contains__(self, key: Any) -> bool:
        with suppress(Exception):
            key = self._key_resolver(key)
        return key in self._values

    @override
    def __getitem__(self, key: KeyT | ResolvableKeyT) -> ValueT:
        return self._values[self._key_resolver(key)]

    @override
    def keys(self) -> Iterable[KeyT]:
        return self._values.keys()


@final
class KeyedCollectionAdapter[KeyT, ResolvableKeyT, ValueT](
    _KeyedCollectionAdapter[KeyT, ResolvableKeyT, ValueT]
):
    """
    A collection of values that are accessible by their primary keys.
    """


@final
class MutableKeyedCollectionAdapter[KeyT, ResolvableKeyT, ValueT, ResolvableValueT](
    _KeyedCollectionAdapter[KeyT, ResolvableKeyT, ValueT],
    MutableKeyedCollection[KeyT, ResolvableKeyT, ValueT, ResolvableValueT],
):
    """
    A mutable ordered collection of values that are accessible by their primary keys.
    """

    def __init__(
        self,
        values: Iterable[ResolvableValueT] | None = None,
        *,
        key: Callable[[ValueT], KeyT],
        key_resolver: Callable[[ResolvableKeyT | KeyT], KeyT] = passthrough,
        value_resolver: Callable[[ResolvableValueT | KeyT], ValueT] = passthrough,
    ):
        super().__init__(key_resolver=key_resolver)
        self._key = key
        self._value_resolver = value_resolver
        if values is not None:
            self.add(*values)

    @override
    def remove(self, *keys: KeyT | ResolvableKeyT) -> None:
        for key in keys:
            del self[key]

    @override
    def __delitem__(self, key: KeyT | ResolvableKeyT) -> None:
        del self._values[self._key_resolver(key)]

    @override
    def clear(self) -> None:
        self._values.clear()

    @override
    def add(self, *values: ValueT | ResolvableValueT) -> None:
        for value in map(
            self._value_resolver,  # ty:ignore[invalid-argument-type]
            values,
        ):
            self._values[self._key(value)] = value



"""
Locale data.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, final

from babel import Locale

from betty.assertion import assert_locale
from betty.data import DataDefinition
from betty.locale import resolve_locale, to_language_tag
from betty.locale.localizable.gettext import _
from betty.portable import CallbackPorter
from betty.property import Property

if TYPE_CHECKING:
    from betty.locale.localizable import ResolvableLocalizable


@final
class LocaleDefinition(DataDefinition[Locale]):
    """
    Define a locale (identifier).
    """

    def __init__(self):
        super().__init__(
            cls=Locale,
            label=_("Locale"),
            description=_('An IETF BCP 47 language tag, such as "nl-NL".'),
            porter=CallbackPorter[Locale](assert_locale(), to_language_tag),
        )


@final
class LocaleProperty(Property):
    """
    A property containing a locale.
    """

    def __init__(
        self,
        *,
        label: ResolvableLocalizable | None = None,
        description: ResolvableLocalizable | None = None,
    ):
        super().__init__(
            LocaleDefinition(),
            label=label,
            description=description,
            resolver=resolve_locale,
        )



"""
Enumerated data types.
"""

from __future__ import annotations

from enum import Enum
from typing import TYPE_CHECKING, final

from betty.data import DataDefinition
from betty.portable import CallbackPorter

if TYPE_CHECKING:
    from betty.locale.localizable import ResolvableLocalizable


@final
class EnumDefinition[EnumT: Enum](DataDefinition[EnumT]):
    """
    An enum data definition.
    """

    def __init__(
        self,
        /,
        cls: type[EnumT],
        *,
        label: ResolvableLocalizable,
        description: ResolvableLocalizable | None = None,
    ):
        from betty.assertion import assert_enum

        super().__init__(
            cls=cls,
            label=label,
            description=description,
            porter=CallbackPorter(assert_enum(cls), lambda enum: enum.value),
        )



from __future__ import annotations  # noqa: D100

import asyncio
from typing import TYPE_CHECKING, Self, final, override

from betty.console.command import Command, CommandDefinition, CommandFunction
from betty.console.project import add_project_argument
from betty.locale.localizable.gettext import _
from betty.service.factory import Manufacturable
from betty.service.requirement.app import require_app

if TYPE_CHECKING:
    import argparse

    from betty.app import App
    from betty.project import Project


@final
@CommandDefinition(
    "serve",
    label=_("Serve a generated site"),
    description=_("This will open your web browser."),
)
class Serve(Manufacturable, Command):
    """
    .. plugin:: command:serve.
    """

    def __init__(self, app: App):
        self._app = app

    @override
    @classmethod
    @require_app
    async def new(cls, app: App, /) -> Self:
        return cls(app)

    @override
    async def configure(self, parser: argparse.ArgumentParser) -> CommandFunction:
        return await add_project_argument(parser, self._command_function, self._app)

    async def _command_function(self, project: Project) -> None:
        from betty import serve

        async with (
            project,
            await serve.BuiltinProjectServer.new(project) as server,
        ):
            await server.show()
            while True:
                await asyncio.sleep(999)
